################################################################################
#
# Copyright (c) 2019, the Perspective Authors.
#
# This file is part of the Perspective library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#

# NOTE: This is essentially a clone of the tornado handler tests,
# but using async/await and starlette handler
import random
import pytest

from aiohttp import web
from datetime import datetime

from perspective import (
    Table,
    PerspectiveManager,
    PerspectiveAIOHTTPHandler,
    aiohttp_websocket as websocket,
    PerspectiveError,
)


data = {
    "a": [i for i in range(10)],
    "b": [i * 1.5 for i in range(10)],
    "c": [str(i) for i in range(10)],
    "d": [datetime(2020, 3, i, i, 30, 45) for i in range(1, 11)],
}

MANAGER = PerspectiveManager()


async def websocket_handler(request):
    handler = PerspectiveAIOHTTPHandler(manager=MANAGER, request=request)
    await handler.run()


@pytest.fixture
def app():
    app = web.Application()
    app.router.add_get("/websocket", websocket_handler)
    return app


class TestPerspectiveAIOHTTPHandler(object):
    def setup_method(self):
        """Flush manager state before each test method execution."""
        MANAGER._tables = {}
        MANAGER._views = {}

    async def websocket_client(self, app, aiohttp_client):
        """Connect and initialize a websocket client connection to the
        Perspective aiottp server.
        """
        client = await aiohttp_client(app)
        return await websocket(
            "http://{}:{}/websocket".format(client.host, client.port), client.session
        )

    @pytest.mark.asyncio
    async def test_aiohttp_handler_init_terminate(self, app, aiohttp_client):
        """Using AIOHTTP's websocket client, test the websocket provided by
        PerspectiveAIOHTTPHandler"""
        client = await self.websocket_client(app, aiohttp_client)
        await client.terminate()

    @pytest.mark.asyncio
    async def test_aiohttp_handler_table_method(self, app, aiohttp_client):
        table_name = str(random.random())
        _table = Table(data)
        MANAGER.host_table(table_name, _table)

        client = await self.websocket_client(app, aiohttp_client)
        table = client.open_table(table_name)

        schema = await table.schema()
        size = await table.size()

        assert schema == {
            "a": "integer",
            "b": "float",
            "c": "string",
            "d": "datetime",
        }

        assert size == 10

    @pytest.mark.asyncio
    async def test_aiohttp_handler_make_table(self, app, aiohttp_client):
        client = await self.websocket_client(app, aiohttp_client)
        table = await client.table(data)
        size = await table.size()

        assert size == 10

        table.update(data)

        size2 = await table.size()
        assert size2 == 20

    @pytest.mark.asyncio
    async def test_aiohttp_handler_table_update(self, app, aiohttp_client):
        table_name = str(random.random())
        _table = Table(data)
        MANAGER.host_table(table_name, _table)

        client = await self.websocket_client(app, aiohttp_client)
        table = client.open_table(table_name)
        size = await table.size()

        assert size == 10

        table.update(data)

        size2 = await table.size()
        assert size2 == 20

    @pytest.mark.asyncio
    async def test_aiohttp_handler_table_update_port(
        self, app, aiohttp_client, sentinel
    ):
        table_name = str(random.random())
        _table = Table(data)
        MANAGER.host_table(table_name, _table)

        client = await self.websocket_client(app, aiohttp_client)
        table = client.open_table(table_name)
        view = await table.view()

        size = await table.size()

        assert size == 10

        for i in range(5):
            await table.make_port()

        port = await table.make_port()

        s = sentinel(False)

        def updater(port_id):
            s.set(True)
            assert port_id == port

        view.on_update(updater)

        table.update(data, port_id=port)

        size2 = await table.size()
        assert size2 == 20
        assert s.get() is True

    @pytest.mark.asyncio
    async def test_aiohttp_handler_table_update_row_delta(
        self, app, aiohttp_client, sentinel
    ):
        table_name = str(random.random())
        _table = Table(data)
        MANAGER.host_table(table_name, _table)

        client = await self.websocket_client(app, aiohttp_client)
        table = client.open_table(table_name)
        view = await table.view()

        size = await table.size()

        assert size == 10

        s = sentinel(False)

        def updater(port_id, delta):
            s.set(True)
            t2 = Table(delta)
            assert t2.view().to_dict() == data
            assert port_id == 0

        view.on_update(updater, mode="row")

        table.update(data)

        size2 = await table.size()
        assert size2 == 20
        assert s.get() is True

    @pytest.mark.asyncio
    async def test_aiohttp_handler_table_update_row_delta_port(
        self, app, aiohttp_client, sentinel
    ):
        table_name = str(random.random())
        _table = Table(data)
        MANAGER.host_table(table_name, _table)

        client = await self.websocket_client(app, aiohttp_client)
        table = client.open_table(table_name)
        view = await table.view()

        size = await table.size()

        assert size == 10

        for i in range(5):
            await table.make_port()

        port = await table.make_port()

        s = sentinel(False)

        def updater(port_id, delta):
            s.set(True)
            t2 = Table(delta)
            assert t2.view().to_dict() == data
            assert port_id == port

        view.on_update(updater, mode="row")

        table.update(data, port_id=port)

        size2 = await table.size()
        assert size2 == 20
        assert s.get() is True

    @pytest.mark.asyncio
    async def test_aiohttp_handler_table_remove(self, app, aiohttp_client):
        table_name = str(random.random())
        _table = Table(data, index="a")
        MANAGER.host_table(table_name, _table)

        client = await self.websocket_client(app, aiohttp_client)
        table = client.open_table(table_name)
        size = await table.size()

        assert size == 10

        table.remove([i for i in range(5)])

        view = await table.view(columns=["a"])
        output = await view.to_dict()

        assert output == {"a": [i for i in range(5, 10)]}

    @pytest.mark.asyncio
    async def test_aiohttp_handler_create_view(self, app, aiohttp_client):
        table_name = str(random.random())
        _table = Table(data)
        MANAGER.host_table(table_name, _table)

        client = await self.websocket_client(app, aiohttp_client)
        table = client.open_table(table_name)
        view = await table.view(columns=["a"])
        output = await view.to_dict()

        assert output == {
            "a": [i for i in range(10)],
        }

    @pytest.mark.asyncio
    async def test_aiohttp_handler_create_view_errors(self, app, aiohttp_client):
        table_name = str(random.random())
        _table = Table(data)
        MANAGER.host_table(table_name, _table)

        client = await self.websocket_client(app, aiohttp_client)
        table = client.open_table(table_name)

        with pytest.raises(PerspectiveError) as exc:
            await table.view(columns=["abcde"])

        assert str(exc.value) == "Invalid column 'abcde' found in View columns.\n"

    @pytest.mark.asyncio
    async def test_aiohttp_handler_create_view_to_arrow(self, app, aiohttp_client):
        table_name = str(random.random())
        _table = Table(data)
        MANAGER.host_table(table_name, _table)

        client = await self.websocket_client(app, aiohttp_client)
        table = client.open_table(table_name)
        view = await table.view()
        output = await view.to_arrow()
        expected = await table.schema()

        assert Table(output).schema(as_string=True) == expected

    @pytest.mark.asyncio
    async def test_aiohttp_handler_create_view_to_arrow_update(
        self, app, aiohttp_client
    ):
        table_name = str(random.random())
        _table = Table(data)
        MANAGER.host_table(table_name, _table)

        client = await self.websocket_client(app, aiohttp_client)
        table = client.open_table(table_name)
        view = await table.view()

        output = await view.to_arrow()

        for i in range(10):
            await table.update(output)

        size2 = await table.size()
        assert size2 == 110



################################################################################
#
# Copyright (c) 2019, the Perspective Authors.
#
# This file is part of the Perspective library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#

from tornado import ioloop
from tornado.websocket import websocket_connect

from .websocket import (
    PerspectiveWebsocketClient,
    PerspectiveWebsocketConnection,
    Periodic,
)


class TornadoPeriodic(Periodic):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._ping_callback = ioloop.PeriodicCallback(
            self._callback,
            callback_time=self._interval,
        )

    async def start(self):
        self._ping_callback.start()

    async def stop(self):
        self._ping_callback.stop()


class PerspectiveTornadoWebsocketConnection(PerspectiveWebsocketConnection):
    def __init__(self):
        self._ws = None

    async def connect(self, url, on_message, max_message_size) -> None:
        self._ws = await websocket_connect(
            url,
            on_message_callback=on_message,
            max_message_size=max_message_size,
        )

    def periodic(self, callback, interval) -> Periodic:
        return TornadoPeriodic(callback=callback, interval=interval)

    async def write(self, message, binary=False):
        return await self._ws.write_message(message, binary=binary)

    async def close(self):
        self._ws.close()


class PerspectiveTornadoClient(PerspectiveWebsocketClient):
    def __init__(self):
        """Create a `PerspectiveTornadoClient` that interfaces with a Perspective server over a Websocket"""
        super(PerspectiveTornadoClient, self).__init__(
            PerspectiveTornadoWebsocketConnection()
        )


async def websocket(url):
    """Create a new websocket client at the given `url` using the thread current
    tornado loop."""
    client = PerspectiveTornadoClient()
    await client.connect(url)
    return client



################################################################################
#
# Copyright (c) 2019, the Perspective Authors.
#
# This file is part of the Perspective library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#

################################################################################
#
# Copyright (c) 2019, the Perspective Authors.
#
# This file is part of the Perspective library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#

import pandas as pd
from perspective import Table, set_threadpool_size

set_threadpool_size(1)


class TestThreadPoolOne:
    def test_threadpool_one_does_not_block_view(self):
        t = Table(
            {"id": int, "symbol": str, "valid": bool, "value": int, "value2": int},
            index="id",
        )
        t.update(
            [
                {"id": 1, "symbol": "A", "valid": False, "value": 5, "value2": 15},
                {"id": 2, "symbol": "A", "valid": True, "value": 10, "value2": 20},
            ]
        )

        v = t.view(
            columns=["symbol", "value", "value3"],
            expressions=["""//value3\n"value" + "value2\""""],
        )

        v_agg = t.view(
            columns=["symbol", "value", "value3"],
            expressions=["""//value3\n"value" + "value2\""""],
            group_by=["symbol"],
            aggregates={"symbol": "first", "value": "sum", "value2": "sum"},
        )

        assert v.to_df().to_dict("records") == [
            {"symbol": "A", "value": 5, "value3": 20.0},
            {"symbol": "A", "value": 10, "value3": 30.0},
        ]

        assert v_agg.to_df().to_dict("records") == [
            {"__ROW_PATH__": [], "symbol": "A", "value": 15, "value3": 50.0},
            {"__ROW_PATH__": ["A"], "symbol": "A", "value": 15, "value3": 50.0},
        ]



################################################################################
#
# Copyright (c) 2019, the Perspective Authors.
#
# This file is part of the Perspective library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#


class PerspectiveError(Exception):
    """Raised for issues within Perspective, i.e. illegal operations."""

    pass



################################################################################
#
# Copyright (c) 2019, the Perspective Authors.
#
# This file is part of the Perspective library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#

from enum import Enum


class Sort(Enum):
    """The sort directions available for use in Perspective. Pass these into
    the `sort` argument of `PerspectiveWidget`.

    Examples:
        >>> widget = PerspectiveWidget(data, sort=[["a", Sort.DESC]])
    """

    ASC = "asc"
    ASC_ABS = "asc abs"
    COL_ASC = "col asc"
    COL_ASC_ABS = "col asc abs"
    COL_DESC = "col desc"
    COL_DESC_ABS = "col desc abs"
    DESC = "desc"
    DESC_ABS = "desc abs"
    NONE = "none"

    @staticmethod
    def options():
        return list(c.value for c in Sort)



################################################################################
#
# Copyright (c) 2019, the Perspective Authors.
#
# This file is part of the Perspective library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#

import pytest
import random
import threading
import asyncio

from datetime import datetime

from fastapi import FastAPI, WebSocket
from fastapi.testclient import TestClient

from perspective import (
    Table,
    PerspectiveManager,
    PerspectiveStarletteHandler,
    PerspectiveError,
)

from perspective.client.starlette_test import websocket

data = {
    "a": [i for i in range(10)],
    "b": [i * 1.5 for i in range(10)],
    "c": [str(i) for i in range(10)],
    "d": [datetime(2020, 3, i, i, 30, 45) for i in range(1, 11)],
}

MANAGER = PerspectiveManager()


def perspective_thread(manager):
    psp_loop = asyncio.new_event_loop()
    manager.set_loop_callback(psp_loop.call_soon_threadsafe)
    psp_loop.run_forever()


thread = threading.Thread(target=perspective_thread, args=(MANAGER,))
thread.daemon = True
thread.start()


async def websocket_handler(websocket: WebSocket):
    handler = PerspectiveStarletteHandler(manager=MANAGER, websocket=websocket)
    await handler.run()


APPLICATION = FastAPI()
APPLICATION.add_api_websocket_route("/websocket", websocket_handler)

CLIENT = TestClient(APPLICATION)


class TestPerspectiveStarletteHandlerAsyncMode(object):
    def setup_method(self):
        """Flush manager state before each test method execution."""
        MANAGER._tables = {}
        MANAGER._views = {}

    async def websocket_client(self):
        return await websocket(CLIENT, "/websocket")

    @pytest.mark.asyncio
    async def test_starlette_handler_async_manager_thread(self):
        table_name = str(random.random())
        _table = Table(data)
        MANAGER.host_table(table_name, _table)
        client = await self.websocket_client()
        table = client.open_table(table_name)
        view = await table.view()
        reqs = []
        for x in range(10):
            reqs.append(table.update(data))
            reqs.append(view.to_arrow())

        await asyncio.gather(*reqs)
        records = await view.to_records()
        assert len(records) == 110

        await client.terminate()



################################################################################
#
# Copyright (c) 2019, the Perspective Authors.
#
# This file is part of the Perspective library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#

import json
from datetime import date, datetime
from functools import partial

import numpy
import pandas
from ipywidgets import DOMWidget
from traitlets import Unicode, observe

from ..core._version import __version__
from ..core.data import deconstruct_pandas
from ..core.exception import PerspectiveError
from ..libpsp import is_libpsp
from ..viewer import PerspectiveViewer


def _type_to_string(t):
    """Convert a type object to a string representing a Perspective-supported
    type.  Redefine here as we can't have any dependencies on libbinding in
    client mode.
    """
    if t is int:
        return "integer"
    elif t is float:
        return "float"
    elif t is bool:
        return "boolean"
    elif t is date:
        return "date"
    elif t is datetime:
        return "datetime"
    elif t is bytes or t is str:
        return "string"
    else:
        raise PerspectiveError(
            "Unsupported type `{0}` in schema - Perspective supports `int`, `float`, `bool`, `date`, `datetime`, and `str` (or `unicode`).".format(
                str(t)
            )
        )


def _serialize_datetime(values):
    """For a list of values, stringify the `date` and `datetime` values
    using strftime."""
    cleaned = []

    for v in values:
        if type(v) is datetime:
            # isinstance(datetime, date) will always return True, so
            # do a type comparison instead.
            cleaned.append(v.strftime("%Y-%m-%d %H:%M:%S"))
        elif type(v) is date:
            cleaned.append(v.strftime("%Y-%m-%d"))
        else:
            cleaned.append(v)

    return cleaned


def _serialize(data):
    """In client mode, PerspectiveWidget will normalize the data before
    passing it to the front-end.

    Widget comms use a custom serializer and messages are not manually
    serialized to JSON by the user, so take special care to remove certain
    unparsable values such as `datetime.date`."""
    if isinstance(data, list):
        # Check if any values are `datetime.date`
        for row in data:
            if not isinstance(row, dict):
                raise PerspectiveError(
                    "Received {} in list dataset, expected `dict`!".format(type(row))
                )

            for k in row.keys():
                if type(row[k]) is datetime:
                    row[k] = row[k].strftime("%Y-%m-%d %H:%M:%S.%f")
                elif type(row[k]) is date:
                    row[k] = row[k].strftime("%Y-%m-%d")
        return data
    elif isinstance(data, dict):
        formatted = data

        for v in data.values():
            if isinstance(v, type):
                # serialize schema values to string
                return {
                    column_name: _type_to_string(data[column_name])
                    for column_name in data
                }
            elif isinstance(v, numpy.ndarray):
                # Convert dicts of numpy arrays to dicts of lists
                formatted = {
                    column_name: data[column_name].tolist() for column_name in data
                }
                break

        for column_name in formatted.keys():
            # Replace `datetime.datetime` and `datetime.date` with string
            formatted[column_name] = _serialize_datetime(formatted[column_name])

        return formatted
    elif isinstance(data, numpy.ndarray):
        # structured or record array
        if not isinstance(data.dtype.names, tuple):
            raise NotImplementedError(
                "Data should be dict of numpy.ndarray or a structured array."
            )

        columns = [data[col].tolist() for col in data.dtype.names]
        formatted = dict(zip(data.dtype.names, columns))

        for column_name in formatted.keys():
            # Replace `datetime.datetime` and `datetime.date` with string
            formatted[column_name] = _serialize_datetime(formatted[column_name])

        return formatted
    elif isinstance(data, pandas.DataFrame) or isinstance(data, pandas.Series):
        # Take flattened dataframe and make it serializable
        d = {}

        for name in data.columns:
            column = data[name]
            values = column.values

            # Timezone-aware datetime64 dtypes throw an exception when using
            # `numpy.issubdtype` - match strings here instead.
            str_dtype = str(column.dtype)
            if "datetime64" in str_dtype:
                # Convert all datetimes to string for serializing
                d[name] = column.dt.strftime("%Y-%m-%d %H:%M:%S").tolist()
            elif str_dtype == "object":
                # Replace `datetime.datetime` and `datetime.date` with string
                d[name] = _serialize_datetime(values)
            else:
                d[name] = values.tolist()
        return d
    else:
        raise NotImplementedError(
            "Cannot serialize a dataset of `{0}`.".format(str(type(data)))
        )


class _PerspectiveWidgetMessage(object):
    """A custom message that will be passed from the Python widget to the
    front-end.

    When creating new messages, use this class as it defines a concrete schema
    for the message and prevents loosely creating `dict` objects everywhere.
    Use `to_dict()` to obtain the message in a form that can be sent through
    IPyWidgets.
    """

    def __init__(self, msg_id, msg_type, msg_data):
        """Create a new PerspectiveWidgetMessage."""
        self.id = msg_id
        self.type = msg_type
        self.data = msg_data

    def to_dict(self):
        """Returns a dictionary representation of the message."""
        return {"id": self.id, "type": self.type, "data": self.data}


class PerspectiveWidget(DOMWidget, PerspectiveViewer):
    """:class`~perspective.PerspectiveWidget` allows for Perspective to be used
    in the form of a JupyterLab IPython widget.

    Using `perspective.Table`, you can create a widget that extends the full
    functionality of `perspective-viewer`.  Changes on the viewer can be
    programatically set on the :class`~perspective.PerspectiveWidget` instance,
    and state is maintained across page refreshes.

    Examples:
        >>> from perspective import Table, PerspectiveWidget
        >>> data = {
        ...     "a": [1, 2, 3],
        ...     "b": [
        ...         "2019/07/11 7:30PM",
        ...         "2019/07/11 8:30PM",
        ...         "2019/07/11 9:30PM"
        ...     ]
        ... }
        >>> tbl = Table(data, index="a")
        >>> widget = PerspectiveWidget(
        ...     tbl,
        ...     group_by=["a"],
        ...     sort=[["b", "desc"]],
        ...     filter=[["a", ">", 1]]
        ... )
        >>> widget.sort
        [["b", "desc"]]
        >>> widget.sort.append(["a", "asc"])
        >>> widget.sort
        [["b", "desc"], ["a", "asc"]]
        >>> widget.update({"a": [4, 5]}) # Browser UI updates
    """

    # Required by ipywidgets for proper registration of the backend
    _model_name = Unicode("PerspectiveModel").tag(sync=True)
    _model_module = Unicode("@finos/perspective-jupyterlab").tag(sync=True)
    _model_module_version = Unicode("~{}".format(__version__)).tag(sync=True)
    _view_name = Unicode("PerspectiveView").tag(sync=True)
    _view_module = Unicode("@finos/perspective-jupyterlab").tag(sync=True)
    _view_module_version = Unicode("~{}".format(__version__)).tag(sync=True)

    def __init__(
        self,
        data,
        index=None,
        limit=None,
        server=False,
        client=not is_libpsp(),
        **kwargs
    ):
        """Initialize an instance of :class`~perspective.PerspectiveWidget`
        with the given table/data and viewer configuration.

        If a pivoted DataFrame or MultiIndex table is passed in, the widget
        preserves pivots and applies them.  See `PerspectiveViewer.__init__` for
        arguments that transform the view shown in the widget.

        Args:
            data (:obj:`Table`|:obj:`View`|:obj:`dict`|:obj:`list`|:obj:`pandas.DataFrame`|:obj:`bytes`|:obj:`str`): a
                `perspective.Table` instance, a `perspective.View` instance, or
                a dataset to be loaded in the widget.

        Keyword Arguments:
            index (:obj:`str`): A column name to be used as the primary key.
                Ignored if `server` is True.

            limit (:obj:`int`): A upper limit on the number of rows in the Table.
                Cannot be set at the same time as `index`, ignored if `server`
                is True.

            server (:obj:`bool`): Whether to run Perspective in "server-only"
                mode, where the front-end client does not have its own Table,
                and instead reads all data and operations from Python.

            client (:obj:`bool`):  If True, convert the dataset into an Apache Arrow
                binary and create the Table in Javascript using a copy of the
                data. Defaults to False.

            kwargs (:obj:`dict`): configuration options for the `PerspectiveViewer`,
                and `Table` constructor if `data` is a dataset.

        Examples:
            >>> widget = PerspectiveWidget(
            ...     {"a": [1, 2, 3]},
            ...     aggregates={"a": "avg"},
            ...     group_by=["a"],
            ...     sort=[["b", "desc"]],
            ...     filter=[["a", ">", 1]],
            ...     expressions=["// new column \n \"Sales\" + \"Profit\""])
        """
        # Trigger special flow when receiving an ArrayBuffer/binary
        self._pending_binary = None

        # If `self.client` is True, the front-end `<perspective-viewer>` is
        # given a copy of the data serialized to JSON, and the Python kernel
        # does not create a `perspective.Table.`
        self.client = client

        # If `self.server` is True, the widget runs in server-only mode where
        # the front-end `<perspective-viewer>` does not create a Table, and
        # will proxy all operations back to the server.
        self.server = server

        # Pass table load options to the front-end, unless in server mode
        self._options = {}

        if index is not None and limit is not None:
            raise PerspectiveError("Index and Limit cannot be set at the same time!")

        # Parse the dataset we pass in - if it's Pandas, preserve pivots
        if isinstance(data, pandas.DataFrame) or isinstance(data, pandas.Series):
            data, config = deconstruct_pandas(data)

            if config.get("group_by", None) and "group_by" not in kwargs:
                kwargs.update({"group_by": config["group_by"]})

            if config.get("split_by", None) and "split_by" not in kwargs:
                kwargs.update({"split_by": config["split_by"]})

            if config.get("columns", None) and "columns" not in kwargs:
                kwargs.update({"columns": config["columns"]})

        # Initialize the viewer
        super(PerspectiveWidget, self).__init__(**kwargs)

        # Handle messages from the the front end
        # `PerspectiveJupyterClient.send()`:
        # - The "data" value of the message should be a JSON-serialized string.
        # - Both `on_msg` and `@observe("value")` must be specified on the
        # handler for custom messages to be parsed by the Python widget.
        self.on_msg(self.handle_message)

        if self.client:
            if is_libpsp():
                from ..libpsp import Table

                if isinstance(data, Table):
                    raise PerspectiveError(
                        "Client mode PerspectiveWidget expects data or schema, not a `perspective.Table`!"
                    )

            if index is not None:
                self._options["index"] = index

            if limit is not None:
                self._options["limit"] = limit

            # cache self._data so creating multiple views don't reserialize the
            # same data
            if not hasattr(self, "_data") or self._data is None:
                self._data = _serialize(data)
        else:
            # If an empty dataset is provided, don't call `load()` and wait
            # for the user to call `load()`.
            if data is None:
                if index is not None or limit is not None:
                    raise PerspectiveError(
                        "Cannot initialize PerspectiveWidget `index` or `limit` without a Table, data, or schema!"
                    )
            else:
                if index is not None:
                    self._options.update({"index": index})

                if limit is not None:
                    self._options.update({"limit": limit})

                self.load(data, **self._options)

    def load(self, data, **options):
        """Load the widget with data. If running in client mode, this method
        serializes the data and calls the browser viewer's load method.
        Otherwise, it calls `Viewer.load()` using `super()`.
        """
        if self.client is True:
            # serialize the data and send a custom message to the browser
            if isinstance(data, pandas.DataFrame) or isinstance(data, pandas.Series):
                data, _ = deconstruct_pandas(data)
            d = _serialize(data)
            self._data = d
        else:
            # Viewer will ignore **options if `data` is a Table or View.
            super(PerspectiveWidget, self).load(data, **options)

        # Notify front-end of load immediately.
        message = self._make_load_message()
        self.send(message.to_dict())

    def update(self, data):
        """Update the widget with new data. If running in client mode, this
        method serializes the data and calls the browser viewer's update
        method. Otherwise, it calls `Viewer.update()` using `super()`.
        """
        if self.client is True:
            # serialize the data and send a custom message to the browser
            if isinstance(data, pandas.DataFrame) or isinstance(data, pandas.Series):
                data, _ = deconstruct_pandas(data)
            d = _serialize(data)
            self.post({"cmd": "update", "data": d})
        else:
            super(PerspectiveWidget, self).update(data)

    def clear(self):
        """Clears the widget's underlying `Table`.

        In client mode, clears the `_data` attribute of the widget.
        """
        if self.client is True:
            self.post({"cmd": "clear"})
            self._data = None
        else:
            super(PerspectiveWidget, self).clear()

    def replace(self, data):
        """Replaces the widget's `Table` with new data conforming to the same
        schema. Does not clear user-set state. If in client mode, serializes
        the data and sends it to the browser.
        """
        if self.client is True:
            if isinstance(data, pandas.DataFrame) or isinstance(data, pandas.Series):
                data, _ = deconstruct_pandas(data)
            d = _serialize(data)
            self.post({"cmd": "replace", "data": d})
            self._data = d
        else:
            super(PerspectiveWidget, self).replace(data)

    def delete(self, delete_table=True):
        """Delete the Widget's data and clears its internal state. If running in
        client mode, sends the `delete()` command to the browser. Otherwise
        calls `delete` on the underlying viewer.

        Args:
            delete_table (`bool`): whether the underlying `Table` will be
                deleted. Defaults to True.
        """
        if self.client is False:
            super(PerspectiveWidget, self).delete(delete_table)
        self.post({"cmd": "delete"})

        # Close the underlying comm and remove widget from the front-end
        self.close()

    def post(self, msg, msg_id=None, binary=False):
        """Post a serialized message to the `PerspectiveJupyterClient`
        in the front end.

        The posted message should conform to the `PerspectiveJupyterMessage`
        interface as defined in `@finos/perspective-jupyterlab`.

        Args:
            msg (dict): a message from `PerspectiveManager` for the front-end
                viewer to process.
            msg_id (int): an integer id that allows the client to process
                the message.
            binary (bool): whether the message contains binary buffers that
                should be sent with a special protocol.
        """
        if binary:
            # The front end will read `buffers` properly.
            self.send(None, buffers=[msg])
        else:
            message = _PerspectiveWidgetMessage(msg_id, "cmd", msg)
            self.send(message.to_dict())

    @observe("value")
    def handle_message(self, widget, content, buffers):
        """Given a message from `PerspectiveJupyterClient.send()`, process the
        message and return the result to `self.post`.

        Args:
            widget: a reference to the `Widget` instance that received the
                message.
            content (dict): the message from the front-end. Automatically
                de-serialized by ipywidgets.
            buffers : optional arraybuffers from the front-end, if any.
        """
        if self._pending_binary:
            msg = self._pending_binary

            # manager looks at the `binary_length` flag so make sure to
            # get rid of it before passing to `_process`.
            del msg["binary_length"]

            # arrow is a `MemoryView` - convert to bytes
            arrow = buffers[0].tobytes()
            msg["args"].insert(0, arrow)

            # Send the message to the manager to process
            post_callback = partial(self.post, msg_id=msg["id"])
            self.manager._process(msg, post_callback)
            self._pending_binary = None
            return

        if content["type"] == "cmd":
            parsed = json.loads(content["data"])

            if parsed["cmd"] == "init":
                self.post({"id": -1, "data": None})
            elif parsed["cmd"] == "table":
                # return the dataset or table name to the front-end
                msg = self._make_load_message()
                self.send(msg.to_dict())
            else:
                # If the message has `binary_length` set, wait for the arrow
                # and join it with the JSON message.
                if parsed.get("binary_length"):
                    self._pending_binary = parsed
                    return

                # For all calls to Perspective, process it in the manager.
                post_callback = partial(self.post, msg_id=parsed["id"])
                self.manager._process(parsed, post_callback)

    def _make_load_message(self):
        """Send a message to the front-end either containing the name of a
        hosted view in Python, so the front-end can create a table in the
        Perspective WebAssembly client, or if `server` is True, the name of a
        Table in python, or the serialized dataset with options if `client`
        is True.

        If the front-end requests data and it has not been loaded yet,
        an error will be logged, and the front-end will wait for `load()` to
        be called, which will notify the front-end of new data.
        """
        msg_data = None

        if self.client and self._data is not None:
            # Send serialized data to the browser, which will run Perspective
            # in client mode: there is no table in the Python kernel.
            msg_data = {"data": self._data, "options": self._options}
        elif self.server and self.table_name is not None:
            # If the `server` kwarg is set during initialization, Perspective
            # will run in server-only mode, where a Table is hosted in the
            # kernel and the front-end proxies pivots, sorts, data requests
            # etc. to the kernel and does not run a Table in the front-end.
            msg_data = {"table_name": self.table_name}
        elif self.table_name is not None:
            # If a view is hosted by the widget's manager (by default),
            # run Perspective in client-server mode: a Table will be created
            # in the front-end that mirrors the Table hosted in the kernel,
            # and updates and edits will be synchronized across the client
            # and the server.
            msg_data = {
                "table_name": self.table_name,
                "options": {},
            }

            index = self.table.get_index()
            limit = self.table.get_limit()

            if index is not None:
                msg_data["options"]["index"] = index
            elif limit is not None:
                msg_data["options"]["limit"] = limit

        if msg_data is not None:
            return _PerspectiveWidgetMessage(-2, "table", msg_data)
        else:
            raise PerspectiveError(
                "Widget does not have any data loaded - use the `load()` method to provide it with data."
            )



################################################################################
#
# Copyright (c) 2019, the Perspective Authors.
#
# This file is part of the Perspective library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#

import datetime
import json
import logging

from collections import deque
from functools import partial

from ..core.exception import PerspectiveError
from ..table import PerspectiveCppError, Table
from ..table._callback_cache import _PerspectiveCallBackCache
from ..table._date_validator import _PerspectiveDateValidator

_date_validator = _PerspectiveDateValidator()


class DateTimeEncoder(json.JSONEncoder):
    """Before sending datetimes over the wire, convert them to Unix timestamps
    in milliseconds since epoch, using Perspective's date validator to
    ensure consistency."""

    def default(self, obj):
        if isinstance(obj, datetime.datetime):
            return _date_validator.to_timestamp(obj)
        else:
            return super(DateTimeEncoder, self).default(obj)


class _PerspectiveManagerInternal(object):

    # Commands that should be blocked from execution when the manager is in
    # `locked` mode, i.e. its tables and views made immutable from remote
    # modification.
    LOCKED_COMMANDS = ["table", "update", "remove", "replace", "clear"]

    def __init__(self, lock=False):
        self._tables = {}
        self._views = {}
        self._callback_cache = _PerspectiveCallBackCache()
        self._queue_process_callback = None
        self._lock = lock

        # Perspective sends binary messages in two messages - a JSON
        # pre-message and the binary itself. Set up flags to handle that
        # special message flow.
        #
        # It is assumed that pre-messages and data messages are ordered:
        # e.g. you can send premessage1 and premessage2, but it is
        # assumed the data will then come as data1 followed by data2
        self._pending_binary = deque()

    def _get_view(self, name):
        """Return a view under management by name."""
        return self._views.get(name, None)

    def _process(self, msg, post_callback, client_id=None):
        """Given a message from the client, process it through the Perspective
        engine.

        Args:
            msg (:obj`dict`): a message from the client with instructions
                that map to engine operations
            post_callback (:obj`callable`): a function that returns data to the
                client. `post_callback` should be a callable that takes two
                parameters: `data` (str), and `binary` (bool), a kwarg that
                specifies whether `data` is a binary string.
        """
        if self._loop_callback:
            self._loop_callback(self.__process, msg, post_callback, client_id)
        else:
            self.__process(msg, post_callback, client_id)

    def __process(self, msg, post_callback, client_id=None):
        """Process the message through the Perspective engine, handling the
        special flow for binary messages along the way."""
        if len(self._pending_binary) >= 0 and isinstance(msg, bytes):
            # The message is an ArrayBuffer, and it needs to be combined with
            # the previous message to reconsitute metadata before processing.
            full_message = self._pending_binary.popleft()
            full_message.pop("binary_length")

            # msg is the binary
            new_args = [msg]

            if len(full_message["args"]) > 1:
                # append additional args, such as port_id. Javascript will
                # serialize the ArrayBuffer as an empty object, so ignore
                # the first argument always.
                new_args += full_message["args"][1:]

            full_message["args"] = new_args
            msg = full_message

        if isinstance(msg, str):
            msg = json.loads(msg)

        if not isinstance(msg, dict):
            raise PerspectiveError(
                "Message passed into `_process` should either be a "
                "JSON-serialized string or a dict."
            )

        if msg.get("binary_length", None):
            # cache the message and wait for the ArrayBuffer that will
            # follow immediately after.
            self._pending_binary.append(msg)
            return

        cmd = msg["cmd"]

        if self._is_locked_command(msg) is True:
            error_string = "`{0}` failed - access denied".format(
                msg["cmd"]
                + (("." + msg["method"]) if msg.get("method", None) is not None else "")
            )
            error_message = self._make_error_message(msg["id"], error_string)
            post_callback(self._message_to_json(msg["id"], error_message))
            return

        try:
            if cmd == "init":
                # The client should wait for the return message on table()
                # and view() to confirm successful creation.
                flags = ["wait_for_response"]

                # Return a message to the client to confirm initialization with
                # a list of feature flags that the client will use to enable
                # or disable behavior depending on its version.
                message = self._make_message(msg["id"], flags)
                post_callback(self._message_to_json(msg["id"], message))
            elif cmd == "table":
                try:
                    # create a new Table and track it
                    data_or_schema = msg["args"][0]
                    self._tables[msg["name"]] = Table(
                        data_or_schema, **msg.get("options", {})
                    )

                    # Return the Table's name to the front end so it can be
                    # resolved.
                    message = self._make_message(msg["id"], msg["name"])
                    post_callback(self._message_to_json(msg["id"], message))
                except IndexError:
                    self._tables[msg["name"]] = []
            elif cmd == "view":
                # Create a new view and notify the client the view has been
                # created by sending the view name back.
                new_view = self._tables[msg["table_name"]].view(**msg.get("config", {}))
                new_view._client_id = client_id
                self._views[msg["view_name"]] = new_view

                # Return the View's name to the front end so it can be
                # resolved.
                message = self._make_message(msg["id"], msg["view_name"])
                post_callback(self._message_to_json(msg["id"], message))
            elif cmd == "table_method" or cmd == "view_method":
                # Call the method on the table/view instance
                self._process_method_call(msg, post_callback, client_id)
        except (PerspectiveError, PerspectiveCppError) as error:
            # Log errors and return them to the client
            error_string = str(error)
            error_message = self._make_error_message(msg["id"], error_string)
            logging.error("[PerspectiveManager._process] %s", error_string)
            post_callback(self._message_to_json(msg["id"], error_message))

    def _process_method_call(self, msg, post_callback, client_id):
        """When the client calls a method, validate the instance it calls on
        and return the result.
        """
        if msg["cmd"] == "table_method":
            table_or_view = self._tables.get(msg["name"], None)
        else:
            table_or_view = self._views.get(msg["name"], None)
            if table_or_view is None:
                error_message = self._make_error_message(
                    msg["id"], "View method cancelled"
                )
                post_callback(self._message_to_json(msg["id"], error_message))
        try:
            if msg.get("subscribe", False) is True:
                self._process_subscribe(msg, table_or_view, post_callback, client_id)
            else:
                # Decide how to dispatch the method
                arguments = {}

                if msg["method"] in (
                    "schema",
                    "expression_schema",
                    "validate_expressions",
                ):
                    # make sure schema returns string types through the
                    # wire API. `as_string` is respected by both the table
                    # and view's `schema` and `expression_schema` methods.
                    arguments["as_string"] = True
                elif msg["method"].startswith("to_"):
                    # parse options in `to_format` calls
                    for d in msg.get("args", []):
                        arguments.update(d)
                else:
                    # Otherwise, arguments are always passed as arrays of
                    # individual arguments.
                    arguments = msg.get("args", [])

                if msg["method"] == "delete":
                    if msg["cmd"] == "view_method":
                        # views can be removed, but tables cannot - intercept
                        # calls to `delete` on the view and return.
                        self._views[msg["name"]].delete()
                        self._views.pop(msg["name"], None)
                        return
                    else:
                        # Return an error when `table.delete()` is called
                        # over the wire API.
                        raise PerspectiveError(
                            "table.delete() cannot be called on a remote table, as the remote has full ownership."
                        )

                # Dispatch the method using the expected argument form
                if msg["method"].startswith("to_"):
                    # to_format takes dictionary of options
                    result = getattr(table_or_view, msg["method"])(**arguments)
                elif msg["method"] in ("update", "remove"):
                    # Apply first arg as positional, then options dict as kwargs
                    data = arguments[0]
                    options = {}
                    if len(arguments) > 1 and isinstance(arguments[1], dict):
                        options = arguments[1]
                    result = getattr(table_or_view, msg["method"])(data, **options)
                elif (
                    msg["cmd"] == "table_method"
                    and msg["method"] == "validate_expressions"
                ):
                    # validate_expressions on the table takes `as_string` in
                    # wargs,
                    result = getattr(table_or_view, msg["method"])(
                        *msg.get("args", []), **arguments
                    )
                else:
                    # otherwise parse arguments as list
                    result = getattr(table_or_view, msg["method"])(*arguments)

                # result has been returned from Perspective, now deliver
                # it back to the user.
                if isinstance(result, bytes) and msg["method"] != "to_csv":
                    # return a binary to the client without JSON serialization,
                    # i.e. when we return an Arrow. If a method is added that
                    # returns a string, this condition needs to be updated as
                    # an Arrow binary is both `str` and `bytes` in Python 2.
                    self._process_bytes(result, msg, post_callback)
                else:
                    # return the result to the client
                    message = self._make_message(msg["id"], result)
                    post_callback(self._message_to_json(msg["id"], message))
        except Exception as error:
            error_string = str(error)
            message = self._make_error_message(msg["id"], error_string)
            logging.error("[PerspectiveManager._process_method_call] %s", error_string)
            post_callback(self._message_to_json(msg["id"], message))

    def _process_subscribe(self, msg, table_or_view, post_callback, client_id):
        """When the client attempts to add or remove a subscription callback,
        validate and perform the requested operation.

        Args:
            msg (dict): the message from the client
            table_or_view {Table|View} : the instance that the subscription
                will be called on.
            post_callback (callable): a method that notifies the client with
                new data.
            client_id (str) : a unique str id that identifies the
                `PerspectiveSession` object that is passing the message.
        """
        try:
            callback = None
            callback_id = msg.get("callback_id", None)
            method = msg.get("method", None)
            args = msg.get("args", [])
            if method and method[:2] == "on":
                # wrap the callback
                callback = partial(self.callback, msg=msg, post_callback=post_callback)
                if callback_id:
                    self._callback_cache.add_callback(
                        {
                            "client_id": client_id,
                            "callback_id": callback_id,
                            "callback": callback,
                            "name": msg.get("name", None),
                        }
                    )
            elif callback_id is not None:
                # pop the callback from the cache of the manager, and
                # remove each of them from the underlying table or view
                popped_callbacks = self._callback_cache.pop_callbacks(callback_id)

                for callback in popped_callbacks:
                    getattr(table_or_view, method)(callback["callback"])
            if callback is not None:
                # call the underlying method on the Table or View, and apply
                # the dictionary of kwargs that is passed through.
                if method == "on_update":
                    # If there are no arguments, make sure we call `on_update`
                    # with mode set to "none".
                    mode = {"mode": "none"}
                    if len(args) > 0:
                        mode = args[0]
                    getattr(table_or_view, method)(callback, **mode)
                else:
                    getattr(table_or_view, method)(callback)
            else:
                logging.info("callback not found for remote call {}".format(msg))
        except Exception as error:
            error_string = str(error)
            message = self._make_error_message(msg["id"], error_string)
            logging.error("[PerspectiveManager._process_subscribe] %s", error_string)
            post_callback(self._message_to_json(msg["id"], message))

    def _process_bytes(self, binary, msg, post_callback):
        """Send a bytestring message to the client without attempting to
        serialize as JSON.

        Perspective's client expects two messages to be sent when a binary
        payload is expected: the first message is a JSON-serialized string with
        the message's `id` and `msg`, and the second message is a bytestring
        without any additional metadata. Implementations of the `post_callback`
        should have an optional kwarg named `binary`, which specifies whether
        `data` is a bytestring or not.

        Args:
            binary (bytes, bytearray) : a byte message to be passed to the client.
            msg (dict) : the original message that generated the binary
                response from Perspective.
            post_callback (callable) : a function that passes data to the
                client, with a `binary` (bool) kwarg that allows it to pass
                byte messages without serializing to JSON.
        """
        # Pass the total length of the binary to the client, so it knows to
        # wait until the arrow has been received in whole.
        msg["binary_length"] = len(binary)

        post_callback(json.dumps(msg, cls=DateTimeEncoder))
        post_callback(binary, binary=True)

    def callback(self, *args, **kwargs):
        """Return a message to the client using the `post_callback` method."""
        orig_msg = kwargs.get("msg")
        id = orig_msg["id"]
        method = orig_msg["method"]
        post_callback = kwargs.get("post_callback")

        if method == "on_update":
            # Coerce the message to be an object so it can be handled in
            # Javascript, where promises cannot be resolved with multiple args.
            updated = {
                "port_id": args[0],
            }
            msg = self._make_message(id, updated)
        else:
            msg = self._make_message(id, None)

        if len(args) > 1 and type(args[1]) == bytes:
            self._process_bytes(args[1], msg, post_callback)
        else:
            post_callback(self._message_to_json(msg["id"], msg))

    def clear_views(self, client_id):
        """Garbage collect views that belong to closed connections."""
        count = 0
        names = []

        if not client_id:
            raise PerspectiveError(
                "Cannot garbage collect views that are not linked to a specific client ID!"
            )

        for name, view in self._views.items():
            if view._client_id == client_id:
                if self._loop_callback:
                    self._loop_callback(view.delete)
                else:
                    view.delete()
                names.append(name)
                count += 1

        for name in names:
            self._views.pop(name)

        if count > 0:
            logging.debug(
                "client {} disconnected - GC {} views in memory".format(
                    client_id, count
                )
            )

    def _make_message(self, id, result):
        """Return a serializable message for a successful result."""
        return {"id": id, "data": result}

    def _make_error_message(self, id, error):
        """Return a serializable message for an error result."""
        return {"id": id, "error": error}

    def _message_to_json(self, id, message):
        """Given a message object to be passed to Perspective, serialize it
        into a string using `DateTimeEncoder` and `allow_nan=False`.

        If an Exception occurs in serialization, catch the Exception and
        return an error message using `self._make_error_message`.

        Args:
            message (:obj:`dict`) a serializable message to be passed to
                Perspective.
        """
        try:
            return json.dumps(message, allow_nan=False, cls=DateTimeEncoder)
        except ValueError as error:
            error_string = str(error)

            # Augment the default error message when invalid values are
            # detected, i.e. `NaN`
            if error_string == "Out of range float values are not JSON compliant":
                error_string = (
                    "Cannot serialize `NaN`, `Infinity` or `-Infinity` to JSON."
                )

            error_message = self._make_error_message(
                id, "JSON serialization error: {}".format(error_string)
            )

            logging.warning(error_message["error"])
            return json.dumps(error_message)

    def _is_locked_command(self, msg):
        """Returns `True` if the manager instance is locked and the command
        is in `_PerspectiveManagerInternal.LOCKED_COMMANDS`, and `False` otherwise."""
        if not self._lock:
            return False

        cmd = msg["cmd"]
        method = msg.get("method", None)

        if cmd == "table_method" and method == "delete":
            # table.delete is blocked, but view.delete is not
            return True

        return cmd == "table" or method in _PerspectiveManagerInternal.LOCKED_COMMANDS



################################################################################
#
# Copyright (c) 2019, the Perspective Authors.
#
# This file is part of the Perspective library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#


class _PerspectiveCallBackCache(object):
    def __init__(self):
        self._callbacks = []

    def add_callback(self, callback):
        self._callbacks.append(callback)

    def remove_callbacks(self, condition):
        """Remove callback functions that satisfy the given condition.

        Args:
            condition (func): a function that returns either True or False. If
                True is returned, filter the item out.
        """
        if not callable(condition):
            raise ValueError("callback filter condition must be a callable function!")
        self._callbacks = [
            callback for callback in self._callbacks if condition(callback) is False
        ]

    def pop_callbacks(self, callback_id):
        """Removes and returns a list of callbacks with the given
        `callback_id`.

        Args:
            callback_id (:obj:`str`) an id that identifies the callback.

        Returns:
            :obj:`list` a list of dicts containing the callbacks that were
                removed.
        """
        popped = []
        new_callbacks = []

        for callback in self._callbacks:
            if callback["callback_id"] == callback_id:
                popped.append(callback)
            else:
                new_callbacks.append(callback)

        return popped

    def get_callbacks(self):
        return self._callbacks

    def __iter__(self):
        return iter(self._callbacks)

    def __len__(self):
        return len(self._callbacks)

    def __repr__(self):
        return str(self._callbacks)



################################################################################
#
# Copyright (c) 2019, the Perspective Authors.
#
# This file is part of the Perspective library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#
import asyncio

from random import random

from ..core.exception import PerspectiveError
from .table_api import PerspectiveTableProxy
from .table_api import table as make_table
from .view_api import PerspectiveViewProxy


class PerspectiveClient(object):
    def __init__(self):
        """A base class for a Perspective client implementation that offers
        a fully async API through an event loop.

        Child classes must implement `send()`, which defines how a message
        should be delivered to the Perspective server implementation.
        """
        self._msg_id = 0
        self._handlers = {}

        # Because we can't reuse a `Future` in Python, we can't just resolve
        # the same `Future` over and over again in order to actually execute
        # on_update/on_delete etc. callbacks.
        #
        # Instead, we store the callbacks on the client and use `callback_id`
        # to access callback references. `self._callback_cache` is a map of
        # callbacks to integer IDs, which is required to allow the `remove_{}`
        # APIs to receive a callable to remove.
        #
        # To execute the callbacks themselves, we can look up the callback
        # in `self._callback_id_cache` (a dict of integer ID to callback),
        # and execute it in the `_handle` method.
        self._callback_cache = {}
        self._callback_id_cache = {}
        self._callback_id = 0

    def open_table(self, name):
        """Return a proxy Table to a `Table` hosted in a server somewhere."""
        return PerspectiveTableProxy(self, name)

    def _handle(self, msg):
        """Given a response from the Perspective server, resolve the Future
        with the response or an exception."""
        if not msg.get("data"):
            return

        handler = self._handlers.get(msg["data"].get("id"))
        if handler:
            future = handler.get("future", None)
            keep_alive = handler.get("keep_alive", False)

            if keep_alive and handler.get("callback_id"):
                # Must look up callback function and execute it, and then
                # return without re-setting the result of the Future.
                callback = self._callback_id_cache.get(handler["callback_id"])
                data = msg["data"]["data"]
                if data and isinstance(data, dict):
                    callback(**data)
                elif data:
                    callback(data)
                else:
                    callback()
                return
            elif future:
                if msg["data"].get("error"):
                    future.set_exception(PerspectiveError(msg["data"]["error"]))
                else:
                    cmd = handler.get("cmd", None)

                    if cmd == "view":
                        future.set_result(
                            PerspectiveViewProxy(self, msg["data"]["data"])
                        )
                    elif cmd == "table":
                        future.set_result(
                            PerspectiveTableProxy(self, msg["data"]["data"])
                        )
                    else:
                        future.set_result(msg["data"]["data"])

            if not keep_alive:
                del self._handlers[msg["data"]["id"]]

    def send(self, msg):
        """Send the message to the Perspective server implementation - must be
        implemented by a child class."""
        raise NotImplementedError()

    def post(self, msg, future=None, keep_alive=False):
        """Given a message and an associated `Future` object, store the future
        and send the message to the server."""
        self._msg_id += 1

        if future:
            handler = {
                "future": future,
            }

            # Handle table and view case here, because we can't reliably set
            # a custom resolve handler like we can do in JS.
            cmd = msg.get("cmd", None)

            if cmd == "view" or cmd == "table":
                handler["cmd"] = cmd

            self._handlers[self._msg_id] = handler
        elif keep_alive:
            handler = {
                "keep_alive": keep_alive,
            }

            if msg.get("callback_id"):
                handler["callback_id"] = msg["callback_id"]

            self._handlers[self._msg_id] = handler

        msg["id"] = self._msg_id

        return asyncio.ensure_future(self.send(msg))

    def table(self, data, index=None, limit=None, name=None):
        """Create a new `Table` in the server implementation, and return
        a proxy to it."""
        name = name or str(random())
        return make_table(self, data, name, index, limit)

    def terminate():
        """Close the connection between the server and client. Must be
        implemented by child classes, although only as part of a public API
        as `terminate()` should only be called by the user."""
        raise NotImplementedError()



################################################################################
#
# Copyright (c) 2019, the Perspective Authors.
#
# This file is part of the Perspective library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#

from random import random


class PerspectiveSession(object):
    """Encapsulates the actions and resources of a single connection to
    Perspective.
    """

    def __init__(self, manager):
        """Create a new session object, keeping track of a unique client_id and
        the manager the session belongs to.

        `PerspectiveSession` should not be constructed directly. Instead, use
        `PerspectiveManager.new_session()` to create a new session.
        """
        self.client_id = str(random())
        self.manager = manager

    def process(self, message, post_callback):
        """Pass a message to the manager's `process` method, which passes the
        result to `post_callback`.

        Additionally, intercept calls to `on_update` in order
        to track the callbacks that were created through this session. This
        allows for cleaning up callbacks after a session closes, which may not
        necessarily involve calling `delete()` on associated views.
        """
        self.manager._process(message, post_callback, client_id=self.client_id)

    def close(self):
        """Remove the views and callbacks that were created within this session
        when the session ends.
        """
        self.manager.clear_views(self.client_id)
        self._clear_callbacks()

    def _clear_callbacks(self):
        # remove all callbacks from the view's cache
        for cb in self.manager._callback_cache:
            if cb["client_id"] == self.client_id:
                view = self.manager._get_view(cb["name"])
                if view:
                    view.remove_update(cb["callback"])
        # remove all callbacks from the manager's cache
        self.manager._callback_cache.remove_callbacks(
            lambda cb: cb["client_id"] == self.client_id
        )



################################################################################
#
# Copyright (c) 2019, the Perspective Authors.
#
# This file is part of the Perspective library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#

from datetime import date, datetime

from ..core.exception import PerspectiveError
from ._accessor import _PerspectiveAccessor
from ._callback_cache import _PerspectiveCallBackCache
from ._date_validator import _PerspectiveDateValidator
from ._state import _PerspectiveStateManager
from ._utils import (
    _dtype_to_pythontype,
    _dtype_to_str,
    _parse_expression_strings,
    _str_to_pythontype,
)
from .libbinding import (
    make_table,
    str_to_filter_op,
    t_dtype,
    t_filter_op,
    t_op,
    validate_expressions,
)
from .view import View


class Table(object):
    def __init__(self, data, limit=None, index=None):
        """Construct a :class:`~perspective.Table` using the provided data or
        schema and optional configuration dictionary.

        :class:`~perspective.Table` instances are immutable - column names and
        data types cannot be changed after creation. If a schema is provided,
        the :class:`~perspective.Table` will be empty.  Subsequent updates MUST
        conform to the column names and data types provided in the schema.

        Args:
            data (:obj:`dict`/:obj:`list`/:obj:`pandas.DataFrame`): Data or
                schema which initializes the :class:`~perspective.Table`.

        Keyword Args:
            index (:obj:`str`): A string column name to use as the
                :class:`~perspective.Table` primary key.
            limit (:obj:`int`): The maximum number of rows the
                :class:`~perspective.Table` should have.  Cannot be set at the
                same time as ``index``. Updates past the limit will begin
                writing at row 0.
        """
        self._is_arrow = isinstance(data, (bytes, bytearray))
        self._is_csv = isinstance(data, str)

        if self._is_csv:
            # CSVs are loaded using the Arrow interface, so is_arrow is
            # always true if we are loading a CSV
            self._is_arrow = True

            if len(data) > 0 and data[0] == ",":
                data = "_" + data

        if self._is_arrow:
            _accessor = data
        else:
            _accessor = _PerspectiveAccessor(data)

        self._date_validator = _PerspectiveDateValidator()

        self._limit = limit
        self._index = index

        # C++ make_table does not accept `None`, so pass in defaults of ""
        # for `index` and 4294967295 for `limit`, but always store `self._index`
        # and `self._limit` as user-provided kwargs or `None`.
        self._table = make_table(
            None,
            _accessor,
            self._limit or 4294967295,
            self._index or "",
            t_op.OP_INSERT,
            False,
            self._is_arrow,
            self._is_csv,
            0,
        )

        self._gnode_id = self._table.get_gnode().get_id()
        self._update_callbacks = _PerspectiveCallBackCache()
        self._delete_callbacks = _PerspectiveCallBackCache()
        self._views = []
        self._delete_callback = None

        pool = self._table.get_pool()
        pool.set_update_delegate(self)
        pool._process()

        # Each table always contains its own instance of state manager.
        self._state_manager = _PerspectiveStateManager()

    def make_port(self):
        """Create a new input port on the underlying `gnode`, and return an
        :obj:`int` containing the ID of the new input port.
        """
        return self._table.make_port()

    def remove_port(self, port_id):
        """Remove the specified port from the underlying `gnode`."""
        self._table.remove_port()

    def get_index(self):
        """Returns the Table's index column, or ``None`` if an index is not
        specified by the user."""
        return self._index

    def get_limit(self):
        """Returns the Table's limit, or ``None`` if an index is not
        specified by the user."""
        return self._limit

    def clear(self):
        """Removes all the rows in the :class:`~perspective.Table`, but
        preserves everything else including the schema and any callbacks or
        registered :class:`~perspective.View`.
        """
        self._state_manager.remove_process(self._table.get_id())
        self._table.reset_gnode(self._gnode_id)

    def replace(self, data):
        """Replaces all rows in the :class:`~perspective.Table` with the new
        data that conforms to the :class:`~perspective.Table` schema.

        Args:
            data (:obj:`dict`/:obj:`list`/:obj:`pandas.DataFrame`): New data
                that will be filled into the :class:`~perspective.Table`.
        """
        self._state_manager.remove_process(self._table.get_id())
        self._table.reset_gnode(self._gnode_id)
        self.update(data)
        self._state_manager.call_process(self._table.get_id())

    def size(self):
        """Returns the row count of the :class:`~perspective.Table`."""
        self._state_manager.call_process(self._table.get_id())
        return self._table.size()

    def schema(self, as_string=False):
        """Returns the schema of this :class:`~perspective.Table`, a :obj:`dict`
        mapping of string column names to python data types.

        Keyword Args:
            as_string (:obj:`bool`): returns the data types as string
                representations, if True

        Returns:
            :obj:`dict`: A key-value mapping of column names to data types.
        """
        s = self._table.get_schema()
        columns = s.columns()
        types = s.types()
        schema = {}
        for i in range(0, len(columns)):
            if columns[i] != "psp_okey":
                if as_string:
                    schema[columns[i]] = _dtype_to_str(types[i])
                else:
                    schema[columns[i]] = _dtype_to_pythontype(types[i])
        return schema

    def validate_expressions(self, expressions, as_string=False):
        """Returns an :obj:`dict` with two keys: "expression_schema", which is
        a schema containing the column names and data types for each valid
        expression in ``expressions``, and "errors", which is a dict of
        expressions to error objects that contain additional metadata:
        `error_message`, `line`, and `column`.

        Args:
            expressions (:obj:`list`): A list of string expressions to validate
                and create a schema from.

        Keyword Args:
            as_string (:obj:`bool`): If True, returns the data types as string
                representations so they can be serialized.
        """
        validated = {"expression_schema": {}, "errors": {}, "expression_alias": {}}

        if len(expressions) == 0:
            return validated

        expressions = _parse_expression_strings(expressions)
        validation_results = validate_expressions(self._table, expressions)
        expression_schema = validation_results.get_expression_schema()
        expression_errors = validation_results.get_expression_errors()

        for expression in expressions:
            # expression_alias can be used to map the alias to the
            # full expression string in the UI.
            validated["expression_alias"][expression[0]] = expression[1]

        for (alias, dtype) in expression_schema.items():
            if not as_string:
                dtype = _str_to_pythontype(dtype)

            validated["expression_schema"][alias] = expression_schema[alias]

        for (alias, error) in expression_errors.items():
            error_dict = {}
            error_dict["error_message"] = error.error_message
            error_dict["line"] = error.line
            error_dict["column"] = error.column

            validated["errors"][alias] = error_dict

        return validated

    def columns(self):
        """Returns the column names of this :class:`~perspective.Table`.

        Returns:
            :obj:`list`: a list of string column names
        """
        return [
            name for name in self._table.get_schema().columns() if name != "psp_okey"
        ]

    def is_valid_filter(self, filter):
        """Tests whether a given filter expression string is valid, e.g. that
        the filter term is not None or an unparsable date/datetime.  `null`/
        `not null` operators don't need a comparison value.

        Args:
            filter (:obj:`string`): The filter expression to validate.

        Returns:
            :obj:`bool`: Whether this filter is valid.
        """
        if isinstance(filter[1], str):
            filter_op = str_to_filter_op(filter[1])
        else:
            filter_op = filter[1]

        if (
            filter_op == t_filter_op.FILTER_OP_IS_NULL
            or filter_op == t_filter_op.FILTER_OP_IS_NOT_NULL
        ):
            # null/not null operators don't need a comparison value
            return True

        value = filter[2]

        if value is None:
            return False

        schema = self.schema()
        in_schema = schema.get(filter[0], None)
        if in_schema and (schema[filter[0]] == date or schema[filter[0]] == datetime):
            if isinstance(value, str):
                value = self._date_validator.parse(value)

        return value is not None

    def update(self, data, port_id=0):
        """Update the :class:`~perspective.Table` with new data.

        Updates on :class:`~perspective.Table` without an explicit ``index``
        are treated as appends.  Updates on :class:`~perspective.Table` with
        an explicit ``index`` should have the index as part of the ``data``
        param, as this instructs the engine to locate the indexed row and write
        into it.  If an index is not provided, the update is treated as an
        append.

        Args:
            data (:obj:`dict`/:obj:`list`/:obj:`pandas.DataFrame`): The data
                with which to update the :class:`~perspective.Table`.

        Examples:
            >>> tbl = Table({"a": [1, 2, 3], "b": ["a", "b", "c"]}, index="a")
            >>> tbl.update({"a": [2, 3], "b": ["a", "a"]})
            >>> tbl.view().to_dict()
            {"a": [1, 2, 3], "b": ["a", "a", "a"]}
        """
        if not port_id:
            port_id = 0

        _is_arrow = isinstance(data, (bytes, bytearray))
        _is_csv = isinstance(data, str)

        if _is_csv:
            # CSVs are loaded using the Arrow interface, so is_arrow is
            # always true if we are loading a CSV
            _is_arrow = True

            if len(data) > 0 and data[0] == ",":
                data = "_" + data

        if _is_arrow:
            _accessor = data
            self._table = make_table(
                self._table,
                _accessor,
                self._limit or 4294967295,
                self._index or "",
                t_op.OP_INSERT,
                True,
                _is_arrow,
                _is_csv,
                port_id,
            )
            self._state_manager.set_process(
                self._table.get_pool(), self._table.get_id()
            )
            return

        columns = self.columns()
        types = self._table.get_schema().types()
        _accessor = _PerspectiveAccessor(data)
        _accessor._names = columns + [
            name for name in _accessor._names if name == "__INDEX__"
        ]
        _accessor._types = types[: len(columns)]

        if _accessor._is_numpy:
            # Try to cast arrays to the Perspective dtype before they end up in
            # the C++, thus allowing for int/floats to be copied when they are
            # the same bit width
            _accessor.try_cast_numpy_arrays()

        if "__INDEX__" in _accessor._names:
            if self._index is not None:
                index_pos = _accessor._names.index(self._index)
                index_dtype = _accessor._types[index_pos]
                _accessor._types.append(index_dtype)
            else:
                _accessor._types.append(t_dtype.DTYPE_INT32)

        self._table = make_table(
            self._table,
            _accessor,
            self._limit or 4294967295,
            self._index or "",
            t_op.OP_INSERT,
            True,
            False,
            False,
            port_id,
        )
        self._state_manager.set_process(self._table.get_pool(), self._table.get_id())

    def remove(self, pkeys, port_id=0):
        """Removes the rows with the primary keys specified in ``pkeys``.

        If the :class:`~perspective.Table` does not have an index, ``remove()``
        has no effect.  Removes propagate to views derived from the table.

        Args:
            pkeys (:obj:`list`): a list of primary keys to indicate the rows
                that should be removed.

        Example:
            >>> tbl = Table({"a": [1, 2, 3]}, index="a")
            >>> tbl.remove([2, 3])
            >>> tbl.view().to_records()
            [{"a": 1}]
        """
        if self._index is None:
            return

        pkeys = list(map(lambda idx: {self._index: idx}, pkeys))
        types = [self._table.get_schema().get_dtype(self._index)]
        _accessor = _PerspectiveAccessor(pkeys)
        _accessor._names = [self._index]
        _accessor._types = types

        t = make_table(
            self._table,
            _accessor,
            self._limit or 4294967295,
            self._index or "",
            t_op.OP_DELETE,
            True,
            False,
            False,
            port_id,
        )

        self._state_manager.set_process(t.get_pool(), t.get_id())

    def view(
        self,
        columns=None,
        group_by=None,
        split_by=None,
        aggregates=None,
        sort=None,
        filter=None,
        expressions=None,
    ):
        """Create a new :class:`~perspective.View` from this
        :class:`~perspective.Table` via the supplied keyword arguments.

        A View is an immutable set of transformations applied to the data stored
        in a :class:`~perspective.Table`, which can be used for querying,
        pivoting, aggregating, sorting, and filtering.

        Keyword Arguments:
            columns (:obj:`list` of :obj:`str`): A list of column names to be
                visible to the user.
            group_by (:obj:`list` of :obj:`str`): A list of column names to
                use as group by.
            split_by (:obj:`list` of :obj:`str`): A list of column names
                to use as split by.
            aggregates (:obj:`dict` of :obj:`str` to :obj:`str`):  A dictionary
                of column names to aggregate types, which specify aggregates
                for individual columns.
            sort (:obj:`list` of :obj:`list` of :obj:`str`): A list of lists,
                each list containing a column name and a sort direction
                (``asc``, ``desc``, ``asc abs``, ``desc abs``, ``col asc``,
                ``col desc``, ``col asc abs``, ``col desc abs``).
            filter (:obj:`list` of :obj:`list` of :obj:`str`):  A list of lists,
                each list containing a column name, a filter comparator, and a
                value to filter by.

        Returns:
            :class:`~perspective.View`: A new :class:`~perspective.View`
                instance bound to this :class:`~perspective.Table`.

        Examples:
            >>> tbl = Table({"a": [1, 2, 3]})
            >>> view = tbl.view(filter=[["a", "==", 1]]
            >>> view.to_dict()
            >>> {"a": [1]}
        """
        self._state_manager.call_process(self._table.get_id())

        config = {}

        if expressions is not None:
            config["expressions"] = _parse_expression_strings(expressions)

        if columns is None:
            config["columns"] = self.columns()

            if expressions is not None:
                for expression in config["expressions"]:
                    config["columns"].append(expression[0])
        else:
            config["columns"] = columns

        if group_by is not None:
            config["group_by"] = group_by
        if split_by is not None:
            config["split_by"] = split_by
        if aggregates is not None:
            config["aggregates"] = aggregates
        if sort is not None:
            config["sort"] = sort
        if filter is not None:
            config["filter"] = filter

        view = View(self, **config)
        self._views.append(view._name)
        return view

    def on_delete(self, callback):
        """Register a callback to be invoked when the
        :func:`~perspective.Table.delete()` method is called on this
        :class:`~perspective.Table`.

        Args:
            callback (:obj:`func`): A callback function to invoke on delete.

        Examples:
            >>> def deleter():
            ...     print("Delete called!")
            >>> table.on_delete(deleter)
            >>> table.delete()
            >>> Delete called!
        """
        if not callable(callback):
            raise ValueError("on_delete callback must be a callable function!")
        self._delete_callbacks.add_callback(callback)

    def remove_delete(self, callback):
        """De-register the supplied callback from the
        :func:`~perspective.Table.delete()` event for this
        :class:`~perspective.Table`

        Examples:
            >>> def deleter():
            ...     print("Delete called!")
            >>> table.on_delete(deleter)
            >>> table.remove_delete(deleter)
            >>> table.delete()
        """
        self._state_manager.remove_process(self._table.get_id())
        if not callable(callback):
            return ValueError("remove_delete callback should be a callable function!")
        self._delete_callbacks.remove_callbacks(lambda cb: cb == callback)

    def delete(self):
        """Delete this :class:`~perspective.Table` and clean up associated
        resources, assuming it has no :class:`~perspective.View` instances
        registered to it (which must be deleted first).
        """
        if len(self._views) > 0:
            raise PerspectiveError(
                "Cannot delete a Table with active views still linked to it "
                + "- call delete() on each view, and try again."
            )
        self._state_manager.remove_process(self._table.get_id())
        self._table.unregister_gnode(self._gnode_id)
        [cb() for cb in self._delete_callbacks]
        self._table = None

    def _update_callback(self, port_id):
        """After `process` completes internally, this method is called by the
        C++ with a `port_id`, indicating the port on which the update was
        processed.

        Arguments:
            port_id (:obj:`int`): an int indicating which port the update
            came from.
        """
        cache = {}
        for callback in self._update_callbacks:
            callback["callback"](port_id=port_id, cache=cache)



################################################################################
#
# Copyright (c) 2019, the Perspective Authors.
#
# This file is part of the Perspective library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#

import asyncio


def async_queue(client, name, method, cmd, *args, **kwargs):
    """Given a method, command, and expected arguments, create a message to
    send to the Perspective server implementation and `post` it through
    the client."""
    arguments = list(args)

    if len(kwargs) > 0:
        arguments.append(kwargs)

    msg = {
        "cmd": cmd,
        "name": name,
        "method": method,
        "args": arguments,
        "subscribe": False,
    }

    future = asyncio.Future()
    client.post(msg, future)
    return future


def subscribe(client, name, method, cmd, *args, **kwargs):
    """Subscribe to an event that occurs on the Perspective server
    implementation, like `on_update`."""
    arguments = list(args)
    callback = None

    for i in range(len(arguments)):
        if callable(arguments[i]):
            callback = arguments.pop(i)

    if len(kwargs) > 0:
        arguments.append(kwargs)

    # Instead of storing in a global map, store the callbacks on the
    # client itself.
    client._callback_id += 1
    client._callback_cache[callback] = client._callback_id
    client._callback_id_cache[client._callback_id] = callback

    msg = {
        "cmd": cmd,
        "name": name,
        "method": method,
        "args": arguments,
        "subscribe": True,
        "callback_id": client._callback_id,
    }

    client.post(msg, keep_alive=True)


def unsubscribe(client, name, method, cmd, *args, **kwargs):
    """Unsubscribe from an event that occurs on the Perspective server
    implementation, like `remove_update`."""
    arguments = list(args)
    callback = None

    for i in range(len(arguments)):
        if callable(arguments[i]):
            callback = arguments.pop(i)

    if len(kwargs) > 0:
        arguments.append(kwargs)

    callback_id = client._callback_cache.get(callback)
    del client._callback_cache[callback]
    del client._callback_id_cache[callback_id]

    msg = {
        "cmd": cmd,
        "name": name,
        "method": method,
        "args": arguments,
        "subscribe": True,
        "callback_id": callback_id,
    }

    client.post(msg, keep_alive=True)



################################################################################
#
# Copyright (c) 2019, the Perspective Authors.
#
# This file is part of the Perspective library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#

# NOTE: This is essentially a clone of the tornado handler tests,
# but using async/await and starlette handler

import pytest
import random

from datetime import datetime

from fastapi import FastAPI, WebSocket
from fastapi.testclient import TestClient

from perspective import (
    Table,
    PerspectiveManager,
    PerspectiveStarletteHandler,
    PerspectiveError,
)

from perspective.client.starlette_test import websocket


data = {
    "a": [i for i in range(10)],
    "b": [i * 1.5 for i in range(10)],
    "c": [str(i) for i in range(10)],
    "d": [datetime(2020, 3, i, i, 30, 45) for i in range(1, 11)],
}

MANAGER = PerspectiveManager()


async def websocket_handler(websocket: WebSocket):
    handler = PerspectiveStarletteHandler(manager=MANAGER, websocket=websocket)
    await handler.run()


APPLICATION = FastAPI()
APPLICATION.add_api_websocket_route("/websocket", websocket_handler)

CLIENT = TestClient(APPLICATION)


class TestPerspectiveStarletteHandler(object):
    def setup_method(self):
        """Flush manager state before each test method execution."""
        MANAGER._tables = {}
        MANAGER._views = {}

    async def websocket_client(self):
        """Connect and initialize a websocket client connection to the
        Perspective starlette server.
        """
        return await websocket(CLIENT, "/websocket")

    @pytest.mark.asyncio
    async def test_starlette_handler_init_terminate(self):
        """Using FastAPI's builtin test client, test the websocket provided by
        PerspectiveStarletteHandler.

        All test methods must import `app`, `http_client`, and `http_port`,
        otherwise a mysterious timeout will occur."""
        client = await self.websocket_client()
        await client.terminate()

    @pytest.mark.asyncio
    async def test_starlette_handler_table_method(self):
        table_name = str(random.random())
        _table = Table(data)
        MANAGER.host_table(table_name, _table)

        client = await self.websocket_client()
        table = client.open_table(table_name)

        schema = await table.schema()
        size = await table.size()

        assert schema == {
            "a": "integer",
            "b": "float",
            "c": "string",
            "d": "datetime",
        }

        assert size == 10

        await client.terminate()

    @pytest.mark.asyncio
    async def test_starlette_handler_make_table(self):
        client = await self.websocket_client()
        table = await client.table(data)
        size = await table.size()

        assert size == 10

        table.update(data)

        size2 = await table.size()
        assert size2 == 20

        await client.terminate()

    @pytest.mark.asyncio
    async def test_starlette_handler_table_update(self):
        table_name = str(random.random())
        _table = Table(data)
        MANAGER.host_table(table_name, _table)

        client = await self.websocket_client()
        table = client.open_table(table_name)
        size = await table.size()

        assert size == 10

        table.update(data)

        size2 = await table.size()
        assert size2 == 20

        await client.terminate()

    @pytest.mark.asyncio
    async def test_starlette_handler_table_update_port(self, sentinel):
        table_name = str(random.random())
        _table = Table(data)
        MANAGER.host_table(table_name, _table)

        client = await self.websocket_client()
        table = client.open_table(table_name)
        view = await table.view()

        size = await table.size()

        assert size == 10

        for i in range(5):
            await table.make_port()

        port = await table.make_port()

        s = sentinel(False)

        def updater(port_id):
            s.set(True)
            assert port_id == port

        view.on_update(updater)

        table.update(data, port_id=port)

        size2 = await table.size()
        assert size2 == 20
        assert s.get() is True

        await client.terminate()

    @pytest.mark.asyncio
    async def test_starlette_handler_table_update_row_delta(self, sentinel):
        table_name = str(random.random())
        _table = Table(data)
        MANAGER.host_table(table_name, _table)

        client = await self.websocket_client()
        table = client.open_table(table_name)
        view = await table.view()

        size = await table.size()

        assert size == 10

        s = sentinel(False)

        def updater(port_id, delta):
            s.set(True)
            assert port_id == 0
            t2 = Table(delta)
            assert t2.view().to_dict() == data

        view.on_update(updater, mode="row")

        table.update(data)

        size2 = await table.size()
        assert size2 == 20
        assert s.get() is True

        await client.terminate()

    @pytest.mark.asyncio
    async def test_starlette_handler_table_update_row_delta_port(self, sentinel):
        table_name = str(random.random())
        _table = Table(data)
        MANAGER.host_table(table_name, _table)

        client = await self.websocket_client()
        table = client.open_table(table_name)
        view = await table.view()

        size = await table.size()

        assert size == 10

        for i in range(5):
            await table.make_port()

        port = await table.make_port()

        s = sentinel(False)

        def updater(port_id, delta):
            s.set(True)
            assert port_id == port

            t2 = Table(delta)
            assert t2.view().to_dict() == data

        view.on_update(updater, mode="row")

        table.update(data, port_id=port)

        size2 = await table.size()
        assert size2 == 20
        assert s.get() is True

        await client.terminate()

    @pytest.mark.asyncio
    async def test_starlette_handler_table_remove(self):
        table_name = str(random.random())
        _table = Table(data, index="a")
        MANAGER.host_table(table_name, _table)

        client = await self.websocket_client()
        table = client.open_table(table_name)
        size = await table.size()

        assert size == 10

        table.remove([i for i in range(5)])

        view = await table.view(columns=["a"])
        output = await view.to_dict()

        assert output == {"a": [i for i in range(5, 10)]}

        await client.terminate()

    @pytest.mark.asyncio
    async def test_starlette_handler_create_view(self):
        table_name = str(random.random())
        _table = Table(data)
        MANAGER.host_table(table_name, _table)

        client = await self.websocket_client()
        table = client.open_table(table_name)
        view = await table.view(columns=["a"])
        output = await view.to_dict()

        assert output == {
            "a": [i for i in range(10)],
        }

        await client.terminate()

    @pytest.mark.asyncio
    async def test_starlette_handler_create_view_errors(self):
        table_name = str(random.random())
        _table = Table(data)
        MANAGER.host_table(table_name, _table)

        client = await self.websocket_client()
        table = client.open_table(table_name)

        with pytest.raises(PerspectiveError) as exc:
            await table.view(columns=["abcde"])

        assert str(exc.value) == "Invalid column 'abcde' found in View columns.\n"

        await client.terminate()

    @pytest.mark.asyncio
    async def test_starlette_handler_create_view_to_arrow(self):
        table_name = str(random.random())
        _table = Table(data)
        MANAGER.host_table(table_name, _table)

        client = await self.websocket_client()
        table = client.open_table(table_name)
        view = await table.view()
        output = await view.to_arrow()
        expected = await table.schema()

        assert Table(output).schema(as_string=True) == expected

        await client.terminate()

    @pytest.mark.asyncio
    async def test_starlette_handler_create_view_to_arrow_update(self):
        table_name = str(random.random())
        _table = Table(data)
        MANAGER.host_table(table_name, _table)

        client = await self.websocket_client()
        table = client.open_table(table_name)
        view = await table.view()

        output = await view.to_arrow()

        for _ in range(10):
            await table.update(output)

        size2 = await table.size()
        assert size2 == 110

        await client.terminate()



################################################################################
#
# Copyright (c) 2019, the Perspective Authors.
#
# This file is part of the Perspective library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#
from datetime import date, datetime
from functools import partial
from types import MethodType

import numpy as np
from perspective import PerspectiveError, PerspectiveWidget, Table
from pytest import raises


def mock_post(self, msg, msg_id=None, assert_msg=None):
    """Mock the widget's `post()` method so we can introspect the contents."""
    assert msg == assert_msg


class TestWidget:
    def test_widget(self):
        data = {"a": np.arange(0, 50)}
        widget = PerspectiveWidget(data, plugin="X Bar")
        assert widget.plugin == "X Bar"
        load_msg = widget._make_load_message()
        assert load_msg.to_dict() == {
            "id": -2,
            "type": "table",
            "data": {
                "table_name": widget.table_name,
                "options": {}
            }
        }

    def test_widget_indexed(self):
        data = {"a": np.arange(0, 50)}
        widget = PerspectiveWidget(data, plugin="X Bar", index="a")
        assert widget.plugin == "X Bar"
        load_msg = widget._make_load_message()
        assert load_msg.to_dict() == {
            "id": -2,
            "type": "table",
            "data": {
                "table_name": widget.table_name,
                "options": {
                    "index": "a"
                }
            }
        }

    def test_widget_no_data(self):
        widget = PerspectiveWidget(None, plugin="X Bar", group_by=["a"])
        assert widget.plugin == "X Bar"
        assert widget.group_by == ["a"]

    def test_widget_schema(self):
        schema = {
            "a": int,
            "b": float,
            "c": bool,
            "d": date,
            "e": datetime,
            "f": str
        }
        widget = PerspectiveWidget(schema)
        assert widget.table.schema() == schema

    def test_widget_schema_with_index(self):
        widget = PerspectiveWidget({"a": int}, index="a")
        assert widget.table.get_index() == "a"

    def test_widget_schema_with_limit(self):
        widget = PerspectiveWidget({"a": int}, limit=5)
        assert widget.table.get_limit() == 5

    def test_widget_no_data_with_index(self):
        # should fail
        with raises(PerspectiveError):
            PerspectiveWidget(None, index="a")

    def test_widget_no_data_with_limit(self):
        # should fail
        with raises(PerspectiveError):
            PerspectiveWidget(None, limit=5)

    def test_widget_eventual_data(self):
        table = Table({"a": np.arange(0, 50)})
        widget = PerspectiveWidget(None, plugin="X Bar")
        assert widget.plugin == "X Bar"

        with raises(PerspectiveError):
            widget._make_load_message()

        widget.load(table)

        load_msg = widget._make_load_message()
        assert load_msg.to_dict() == {
            "id": -2,
            "type": "table",
            "data": {
                "table_name": widget.table_name,
                "options": {}
            }
        }

    def test_widget_eventual_data_server(self):
        widget = PerspectiveWidget(None, plugin="X Bar", server=True)
        assert widget.plugin == "X Bar"
        widget.load({"a": np.arange(0, 50)}, index="a")
        load_msg = widget._make_load_message()
        assert load_msg.to_dict() == {
            "id": -2,
            "type": "table",
            "data": {
                "table_name": widget.table_name,
            }
        }

    def test_widget_eventual_data_indexed(self):
        widget = PerspectiveWidget(None, plugin="X Bar")
        assert widget.plugin == "X Bar"
        widget.load({"a": np.arange(0, 50)}, index="a")
        load_msg = widget._make_load_message()
        assert load_msg.to_dict() == {
            "id": -2,
            "type": "table",
            "data": {
                "table_name": widget.table_name,
                "options": {
                    "index": "a"
                }
            }
        }

    def test_widget_eventual_table_indexed(self):
        table = Table({"a": np.arange(0, 50)}, index="a")
        widget = PerspectiveWidget(None, plugin="X Bar")
        assert widget.plugin == "X Bar"
        widget.load(table)
        load_msg = widget._make_load_message()
        assert load_msg.to_dict() == {
            "id": -2,
            "type": "table",
            "data": {
                "table_name": widget.table_name,
                "options": {
                    "index": "a"
                }
            }
        }

    def test_widget_load_table(self):
        table = Table({"a": np.arange(0, 50)})
        widget = PerspectiveWidget(table, plugin="X Bar")
        assert widget.plugin == "X Bar"
        load_msg = widget._make_load_message()
        assert load_msg.to_dict() == {
            "id": -2,
            "type": "table",
            "data": {
                "table_name": widget.table_name,
                "options": {}
            }
        }

    def test_widget_load_table_indexed(self):
        table = Table({"a": np.arange(0, 50)}, index="a")
        widget = PerspectiveWidget(table, plugin="X Bar")
        assert widget.plugin == "X Bar"
        load_msg = widget._make_load_message()
        assert load_msg.to_dict() == {
            "id": -2,
            "type": "table",
            "data": {
                "table_name": widget.table_name,
                "options": {
                    "index": "a"
                }
            }
        }

    def test_widget_load_table_ignore_limit(self):
        table = Table({"a": np.arange(0, 50)})
        widget = PerspectiveWidget(table, limit=1)
        assert widget.table.size() == 50

    def test_widget_pass_index(self):
        data = {"a": [1, 2, 3, 1]}
        widget = PerspectiveWidget(data, index="a")
        assert widget.table.size() == 3

    def test_widget_pass_limit(self):
        data = {"a": np.arange(0, 50)}
        widget = PerspectiveWidget(data, limit=1)
        assert widget.table.size() == 1

    def test_widget_pass_options_invalid(self):
        data = {"a": np.arange(0, 50)}
        with raises(PerspectiveError):
            PerspectiveWidget(data, index="index", limit=1)

    # server mode
    def test_widget_load_table_server(self):
        table = Table({"a": np.arange(0, 50)})
        widget = PerspectiveWidget(table, server=True)
        load_msg = widget._make_load_message()
        assert load_msg.to_dict() == {
            "id": -2,
            "type": "table",
            "data": {
                "table_name": widget.table_name
            }
        }

    def test_widget_no_data_with_server(self):
        # should fail
        widget = PerspectiveWidget(None, server=True)
        with raises(PerspectiveError):
            widget._make_load_message()

    def test_widget_eventual_data_with_server(self):
        # should fail
        widget = PerspectiveWidget(None, server=True)

        with raises(PerspectiveError):
            widget._make_load_message()

        # then succeed
        widget.load(Table({"a": np.arange(0, 50)}))
        load_msg = widget._make_load_message()
        assert load_msg.to_dict() == {
            "id": -2,
            "type": "table",
            "data": {
                "table_name": widget.table_name
            }
        }

    # clear

    def test_widget_clear(self):
        data = {"a": np.arange(0, 50)}
        widget = PerspectiveWidget(data)
        widget.clear()
        assert widget.table.size() == 0

    def test_widget_clear_server(self):
        data = {"a": np.arange(0, 50)}
        widget = PerspectiveWidget(data, server=True)
        widget.clear()
        assert widget.table.size() == 0

    # replace

    def test_widget_replace(self):
        data = {"a": np.arange(0, 50)}
        widget = PerspectiveWidget(data)
        widget.replace({"a": [1]})
        assert widget.table.size() == 1

    def test_widget_replace_server(self):
        data = {"a": np.arange(0, 50)}
        widget = PerspectiveWidget(data, server=True)
        widget.replace({"a": [1]})
        assert widget.table.size() == 1

    # delete

    def test_widget_delete(self):
        data = {"a": np.arange(0, 50)}
        widget = PerspectiveWidget(data)
        mocked_post = partial(mock_post, assert_msg={
            "cmd": "delete"
        })
        widget.post = MethodType(mocked_post, widget)
        widget.delete()
        assert widget.table is None

    def test_widget_delete_with_view(self):
        data = {"a": np.arange(0, 50)}
        widget = PerspectiveWidget(data)

        # create a view on the manager
        table_name, table = list(widget.manager._tables.items())[0]
        make_view_message = {"id": 1, "table_name": table_name, "view_name": "view1", "cmd": "view", "config": {"group_by": ["a"]}}
        widget.manager._process(make_view_message, lambda x: True)

        assert len(widget.manager._views) == 1

        mocked_post = partial(mock_post, assert_msg={
            "cmd": "delete"
        })

        widget.post = MethodType(mocked_post, widget)
        widget.delete()

        assert widget.table is None



################################################################################
#
# Copyright (c) 2019, the Perspective Authors.
#
# This file is part of the Perspective library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#

from pytest import raises
from perspective import PerspectiveError, PerspectiveViewer,\
                        PerspectiveWidget, Aggregate


class TestAggregates:

    def test_aggregates_widget_load(self):
        aggs = {
            "a": Aggregate.AVG,
            "b": Aggregate.LAST
        }
        data = {"a": [1, 2, 3], "b": ["a", "b", "c"]}
        widget = PerspectiveWidget(data, aggregates=aggs)
        assert widget.aggregates == aggs

    def test_aggregates_widget_load_weighted_mean(self):
        aggs = {
            "a": Aggregate.AVG,
            "b": ["weighted mean", "a"]
        }
        data = {"a": [1, 2, 3], "b": ["a", "b", "c"]}
        widget = PerspectiveWidget(data, aggregates=aggs)
        assert widget.aggregates == aggs

    def test_aggregates_widget_setattr(self):
        data = {"a": [1, 2, 3], "b": ["a", "b", "c"]}
        widget = PerspectiveWidget(data)
        widget.aggregates = {
            "a": Aggregate.ANY,
            "b": Aggregate.LAST
        }
        assert widget.aggregates == {
            "a": "any",
            "b": "last"
        }

    def test_aggregates_widget_load_invalid(self):
        data = {"a": [1, 2, 3], "b": ["a", "b", "c"]}
        with raises(PerspectiveError):
            PerspectiveWidget(data, aggregates={"a": "?"})

    def test_aggregates_widget_setattr_invalid(self):
        data = {"a": [1, 2, 3], "b": ["a", "b", "c"]}
        widget = PerspectiveWidget(data)
        with raises(PerspectiveError):
            widget.aggregates = {"a": "?"}

    def test_aggregates_widget_init_all(self):
        data = {"a": [1, 2, 3], "b": ["a", "b", "c"]}
        for agg in Aggregate:
            widget = PerspectiveWidget(data, aggregates={"a": agg})
            assert widget.aggregates == {"a": agg.value}

    def test_aggregates_widget_set_all(self):
        data = {"a": [1, 2, 3], "b": ["a", "b", "c"]}
        widget = PerspectiveWidget(data)
        for agg in Aggregate:
            widget.aggregates = {"a": agg}
            assert widget.aggregates == {"a": agg.value}

    def test_aggregates_viewer_load(self):
        viewer = PerspectiveViewer(aggregates={"a": Aggregate.AVG})
        assert viewer.aggregates == {"a": "avg"}

    def test_aggregates_viewer_setattr(self):
        viewer = PerspectiveViewer()
        viewer.aggregates = {"a": Aggregate.AVG}
        assert viewer.aggregates == {"a": "avg"}

    def test_aggregates_viewer_init_all(self):
        for agg in Aggregate:
            viewer = PerspectiveViewer(aggregates={"a": agg})
            assert viewer.aggregates == {"a": agg.value}

    def test_aggregates_viewer_set_all(self):
        viewer = PerspectiveViewer()
        for agg in Aggregate:
            viewer.aggregates = {"a": agg}
            assert viewer.aggregates == {"a": agg.value}



# *****************************************************************************
#
# Copyright (c) 2019, the Perspective Authors.
#
# This file is part of the Perspective library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#
from __future__ import print_function

import io
import os
import os.path
import platform
import re
import subprocess
import sys
import sysconfig
from codecs import open
from distutils.version import LooseVersion

from setuptools import Extension, find_packages, setup
from setuptools.command.build_ext import build_ext
from setuptools.command.sdist import sdist

try:
    from shutil import which

    CPU_COUNT = os.cpu_count()
except ImportError:
    raise Exception("Requires Python 3.7 or later")

here = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(here, "README.md"), encoding="utf-8") as f:
    long_description = f.read().replace("\r\n", "\n")

if sys.version_info.major < 3:
    raise Exception("Requires Python 3.7 or later")


def get_version(file, name="__version__"):
    """Get the version of the package from the given file by
    executing it and extracting the given `name`.
    """
    path = os.path.realpath(file)
    version_ns = {}
    with io.open(path, encoding="utf8") as f:
        exec(f.read(), {}, version_ns)
    return version_ns[name]


version = get_version(os.path.join(here, "perspective", "core", "_version.py"))

requires = [
    "ipywidgets>=7.5.1,<8",
    "future>=0.16.0,<1",
    "numpy>=1.13.1,<2",
    "pandas>=0.22.0,<2",
    "python-dateutil>=2.8.0,<3",
    "traitlets>=4.3.2,<6",
]

requires_aiohttp = ["aiohttp>=3,<4"]

requires_jupyter = ["jupyterlab>=3.2,<4"]

requires_starlette = ["fastapi>=0.70,<1", "starlette>=0.20,<1"]

requires_tornado = ["tornado>=4.5.3,<6.2"]

requires_dev = (
    [
        "black==22.8",
        "Faker>=1.0.0",
        "flake8>=5",
        "flake8-black>=0.3.3",
        "httpx>=0.23,<1",
        "pip",
        "psutil>=5,<6",
        "pybind11>=2.4.0,<3",
        "pyarrow>=0.16.0",
        "pytest>=4.3.0",
        "pytest-aiohttp>=1",
        "pytest-asyncio>=0.20",
        "pytest-cov>=3",
        "pytest-check-links>=0.7",
        "pytest-tornado>=0.8",
        "pytz>=2022",
        "Sphinx>=1.8.4",
        "sphinx-markdown-builder>=0.5.2",
        "wheel",
    ]
    + requires
    + requires_aiohttp
    + requires_jupyter
    + requires_starlette
    + requires_tornado
)


class PSPExtension(Extension):
    def __init__(self, name, sourcedir="dist"):
        Extension.__init__(self, name, sources=[])
        self.sourcedir = os.path.abspath(sourcedir)


class PSPBuild(build_ext):
    def run(self):
        self.run_cmake()

    def run_cmake(self):
        self.cmake_cmd = which("cmake")
        try:
            out = subprocess.check_output([self.cmake_cmd, "--version"])
        except OSError:
            raise RuntimeError(
                "CMake must be installed to build the following extensions: "
                + ", ".join(e.name for e in self.extensions)
            )

        if platform.system() == "Windows":
            cmake_version = LooseVersion(
                re.search(r"version\s*([\d.]+)", out.decode()).group(1)
            )
            if cmake_version < "3.1.0":
                raise RuntimeError("CMake >= 3.1.0 is required on Windows")

        for ext in self.extensions:
            self.build_extension_cmake(ext)

    def build_extension_cmake(self, ext):
        extdir = os.path.abspath(os.path.dirname(self.get_ext_fullpath(ext.name)))
        cfg = "Debug" if self.debug else "Release"

        PYTHON_VERSION = "{}.{}".format(sys.version_info.major, sys.version_info.minor)

        cmake_args = [
            "-DCMAKE_LIBRARY_OUTPUT_DIRECTORY="
            + os.path.abspath(os.path.join(extdir, "perspective", "table")).replace(
                "\\", "/"
            ),
            "-DCMAKE_BUILD_TYPE=" + cfg,
            "-DPSP_CPP_BUILD=1",
            "-DPSP_WASM_BUILD=0",
            "-DPSP_PYTHON_BUILD=1",
            "-DPSP_PYTHON_VERSION={}".format(PYTHON_VERSION),
            "-DPython_ADDITIONAL_VERSIONS={}".format(PYTHON_VERSION),
            "-DPython_FIND_VERSION={}".format(PYTHON_VERSION),
            "-DPython_EXECUTABLE={}".format(sys.executable).replace("\\", "/"),
            "-DPYTHON_LIBRARY={}".format(sysconfig.get_config_var("LIBDIR")).replace(
                "\\", "/"
            ),
            "-DPYTHON_INCLUDE_DIR={}".format(
                sysconfig.get_config_var("INCLUDEPY")
            ).replace("\\", "/"),
            "-DPython_ROOT_DIR={}".format(sys.prefix).replace("\\", "/"),
            "-DPython_ROOT={}".format(sys.prefix).replace("\\", "/"),
            "-DPSP_CMAKE_MODULE_PATH={folder}".format(
                folder=os.path.join(ext.sourcedir, "cmake")
            ).replace("\\", "/"),
            "-DPSP_CPP_SRC={folder}".format(folder=ext.sourcedir).replace("\\", "/"),
            "-DPSP_PYTHON_SRC={folder}".format(
                folder=os.path.join(ext.sourcedir, "..", "perspective").replace(
                    "\\", "/"
                )
            ),
        ]

        build_args = ["--config", cfg]
        env = os.environ.copy()

        if platform.system() == "Windows":
            import distutils.msvccompiler as dm

            # https://wiki.python.org/moin/WindowsCompilers#Microsoft_Visual_C.2B-.2B-_14.0_with_Visual_Studio_2015_.28x86.2C_x64.2C_ARM.29
            msvc = {
                "12": "Visual Studio 12 2013",
                "14": "Visual Studio 14 2015",
                "14.0": "Visual Studio 14 2015",
                "14.1": "Visual Studio 15 2017",
                "14.2": "Visual Studio 16 2019",
                "14.3": "Visual Studio 17 2022",
            }.get(dm.get_build_version(), "Visual Studio 15 2017")

            cmake_args.extend(
                [
                    "-DCMAKE_LIBRARY_OUTPUT_DIRECTORY_{}={}".format(
                        cfg.upper(), extdir
                    ).replace("\\", "/"),
                    "-G",
                    os.environ.get("PSP_GENERATOR", msvc),
                ]
            )

            vcpkg_toolchain_file = os.path.abspath(
                os.environ.get(
                    "PSP_VCPKG_PATH",
                    os.path.join(
                        "..", "..", "vcpkg\\scripts\\buildsystems\\vcpkg.cmake"
                    ),
                )
            )

            if os.path.exists(vcpkg_toolchain_file):
                cmake_args.append(
                    "-DCMAKE_TOOLCHAIN_FILE={}".format(vcpkg_toolchain_file)
                )

            if sys.maxsize > 2**32:
                # build 64 bit to match python
                cmake_args += ["-A", "x64"]

            build_args += [
                "--",
                "/m:{}".format(CPU_COUNT),
                "/p:Configuration={}".format(cfg),
            ]
        else:
            cmake_args += ["-DCMAKE_BUILD_TYPE=" + cfg]
            build_args += [
                "--",
                "-j2"
                if os.environ.get("DOCKER", "")
                else "-j{}".format(env.get("PSP_NUM_CPUS", CPU_COUNT)),
            ]

        env["PSP_ENABLE_PYTHON"] = "1"
        env["OSX_DEPLOYMENT_TARGET"] = os.environ.get(
            "PSP_OSX_DEPLOYMENT_TARGET", "10.9"
        )
        env["MACOSX_DEPLOYMENT_TARGET"] = os.environ.get(
            "PSP_OSX_DEPLOYMENT_TARGET", "10.9"
        )

        if not os.path.exists(self.build_temp):
            os.makedirs(self.build_temp)

        subprocess.check_call(
            [self.cmake_cmd, os.path.abspath(ext.sourcedir)] + cmake_args,
            cwd=self.build_temp,
            env=env,
            stderr=subprocess.STDOUT,
        )
        subprocess.check_call(
            [self.cmake_cmd, "--build", "."] + build_args,
            cwd=self.build_temp,
            env=env,
            stderr=subprocess.STDOUT,
        )
        print()  # Add an empty line for cleaner output


class PSPCheckSDist(sdist):
    def run(self):
        self.run_check()
        super(PSPCheckSDist, self).run()

    def run_check(self):
        for file in ("CMakeLists.txt", "cmake", "src"):
            path = os.path.abspath(os.path.join(here, "dist", file))
            if not os.path.exists(path):
                raise Exception(
                    "Path is missing! {}\nMust run `yarn build_python` before building sdist so cmake files are installed".format(
                        path
                    )
                )


setup(
    name="perspective-python",
    version=version,
    description="Python bindings and JupyterLab integration for Perspective",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/finos/perspective",
    author="Perspective Authors",
    author_email="info@finos.org",
    license="Apache 2.0",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
    keywords="analytics tools plotting",
    packages=find_packages(exclude=["bench", "bench.*"]),
    include_package_data=True,
    zip_safe=False,
    python_requires=">=3.7",
    install_requires=requires,
    extras_require={
        "aiohttp": requires_aiohttp,
        "dev": requires_dev,
        "develop": requires_dev,
        "fastapi": requires_starlette,
        "jupyter": requires_jupyter,
        "starlette": requires_starlette,
        "tornado": requires_tornado,
    },
    ext_modules=[PSPExtension("perspective")],
    cmdclass=dict(build_ext=PSPBuild, sdist=PSPCheckSDist),
)



################################################################################
#
# Copyright (c) 2019, the Perspective Authors.
#
# This file is part of the Perspective library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#

from perspective import Table, set_threadpool_size

def compare_delta(received, expected):
    """Compare an arrow-serialized row delta by constructing a Table."""
    tbl = Table(received)
    assert tbl.view().to_dict() == expected

class TestThreadpool(object):
    def test_set_threadpool_size(self):
        set_threadpool_size(1)
        data = [{"a": 1, "b": 2}, {"a": 3, "b": 4}]
        tbl = Table(data)
        view = tbl.view()
        assert view.num_rows() == 2
        assert view.num_columns() == 2
        assert view.schema() == {
            "a": int,
            "b": int
        }
        assert view.to_records() == data

    def test_set_threadpool_size_max(self):
        set_threadpool_size(None)
        data = [{"a": 1, "b": 2}, {"a": 3, "b": 4}]
        tbl = Table(data)
        view = tbl.view()
        assert view.num_rows() == 2
        assert view.num_columns() == 2
        assert view.schema() == {
            "a": int,
            "b": int
        }
        assert view.to_records() == data



################################################################################
#
# Copyright (c) 2019, the Perspective Authors.
#
# This file is part of the Perspective library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#

import random
from perspective.table import Table

data = {
    "a": [1, 2, 3, 4],
    "b": ["a", "b", "c", "d"],
    "c": [True, False, True, False]
}


class TestPorts(object):

    def test_make_port_sequential(self):
        table = Table(data)
        port_ids = []

        for i in range(10):
            port_ids.append(table.make_port())

        assert port_ids == list(range(1, 11))

    def test_make_port_sequential_and_update(self):
        table = Table(data)
        port_ids = []

        for i in range(10):
            port_ids.append(table.make_port())

        assert port_ids == list(range(1, 11))

        for i in range(1, 11):
            table.update({
                "a": [i],
                "b": ["a"],
                "c": [True]
            }, port_id=i)

        view = table.view()
        result = view.to_dict()

        assert result == {
            "a": [1, 2, 3, 4] + [i for i in range(1, 11)],
            "b": ["a", "b", "c", "d"] + ["a" for i in range(10)],
            "c": [True, False, True, False] + [True for i in range(10)]
        }

    def test_arbitary_port_updates(self):
        table = Table(data)
        port_ids = []

        for i in range(10):
            port_ids.append(table.make_port())

        assert port_ids == list(range(1, 11))

        port = random.randint(0, 10)

        table.update(data, port_id=port)

        assert table.size() == 8

        assert table.view().to_dict() == {
            "a": [1, 2, 3, 4] * 2,
            "b": ["a", "b", "c", "d"] * 2,
            "c": [True, False, True, False] * 2
        }

    def test_ports_should_only_notify_if_they_have_a_queued_update(self):
        table = Table(data)
        port_ids = []

        for i in range(10):
            port_ids.append(table.make_port())

        assert port_ids == list(range(1, 11))

        view = table.view()
        ports_to_update = [random.randint(0, 10) for i in range(5)]

        def callback(port_id):
            assert port_id in ports_to_update

        view.on_update(callback)

        for port in ports_to_update:
            table.update(data, port_id=port)

    def test_ports_should_have_unique_deltas(self):
        table = Table(data)
        port_ids = []

        for i in range(10):
            port_ids.append(table.make_port())

        assert port_ids == list(range(1, 11))

        view = table.view()
        ports_to_update = [random.randint(0, 10) for i in range(5)]
        unique_data = {port: [{"a": port, "b": str(port), "c": True}] for port in ports_to_update}

        def callback(port_id, delta):
            assert port_id in ports_to_update
            _t = Table(delta)
            _v = _t.view()
            assert _v.to_records() == unique_data[port]
            _v.delete()
            _t.delete()

        view.on_update(callback, mode="row")

        for port in ports_to_update:
            table.update(unique_data[port], port_id=port)

    def test_ports_should_queue_updates_properly(self):
        table = Table(data)
        port_ids = []

        for i in range(10):
            port_ids.append(table.make_port())

        assert port_ids == list(range(1, 11))

        view = table.view()
        ports_to_update = [random.randint(0, 10) for i in range(5)]

        def callback(port_id):
            assert port_id in ports_to_update

        view.on_update(callback)

        for port in ports_to_update:
            table.update(data, port_id=port)

    def test_ports_multiple_tables_with_different_ports(self):
        server = Table(data)
        client = Table(data)

        for i in range(random.randint(5, 15)):
            # reserve an arbitary number of ports
            server.make_port()

        # port for client is now far above the ports "ON" the client, as the
        # client ports will begin creation at 1.
        server_port_for_client = server.make_port()
        client_port = client.make_port()

        server_view = server.view()
        client_view = client.view()

        # when the client updates, check whether the port id matches that
        # of the server, and complete the test.
        def client_callback(port_id, delta):
            if port_id == client_port:
                print("UPDATING SERVER")
                server.update(delta, port_id=server_port_for_client)

        # when the server updates, pass the update back to the client
        def server_callback(port_id, delta):
            print("UPDATING CLIENT")
            assert port_id == server_port_for_client
            assert server.size() == 8
            server_view.delete()
            server.delete()
            client_view.delete()
            client.delete()

        client_view.on_update(client_callback, mode="row")
        server_view.on_update(server_callback, mode="row")

        client.update(data, port_id=client_port)



################################################################################
#
# Copyright (c) 2019, the Perspective Authors.
#
# This file is part of the Perspective library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#
import asyncio

from random import random
from functools import partial
from .dispatch import async_queue, subscribe, unsubscribe


def view(
    client,
    table_name,
    columns=None,
    group_by=None,
    split_by=None,
    aggregates=None,
    sort=None,
    filter=None,
    expressions=None,
):
    """Create a new View by posting a message to the Perspective server
    implementation through `client`, returning a Future that will resolve to a
    `PerspectiveViewProxy` object whose API must be called with `await` or
    `yield`, or an Exception if the View creation failed.
    """
    name = str(random())

    config = {
        "columns": columns,
        "group_by": group_by,
        "split_by": split_by,
        "aggregates": aggregates,
        "sort": sort,
        "filter": filter,
        "expressions": expressions,
    }

    msg = {
        "cmd": "view",
        "view_name": name,
        "table_name": table_name,
        "config": config,
    }

    future = asyncio.Future()
    client.post(msg, future)
    return future


class PerspectiveViewProxy(object):
    def __init__(self, client, name):
        """A proxy for a Perspective `View` object elsewhere, i.e. on a remote
        server accessible through a Websocket.

        All public API methods on this proxy are async, and must be called
        with `await` or a `yield`-based coroutine.

        Args:
            client (:obj:`PerspectiveClient`): A `PerspectiveClient` that is
                set up to send messages to a Perspective server implementation
                elsewhere.

            name (:obj:`str`): a `str` name for the View. Automatically
                generated if using the `view` function defined above.
        """
        self._client = client
        self._name = name
        self._async_queue = partial(async_queue, self._client, self._name)
        self._subscribe = partial(subscribe, self._client, self._name)
        self._unsubscribe = partial(unsubscribe, self._client, self._name)

    def get_config(self):
        return self._async_queue("get_config", "view_method")

    def sides(self):
        return self._async_queue("sides", "view_method")

    def num_rows(self):
        return self._async_queue("num_rows", "view_method")

    def num_columns(self):
        return self._async_queue("num_columns", "view_method")

    def get_min_max(self):
        return self._async_queue("get_min_max", "view_method")

    def get_row_expanded(self, idx):
        return self._async_queue("get_row_expanded", "view_method", idx)

    def expand(self, idx):
        return self._async_queue("expand", "view_method", idx)

    def collapse(self, idx):
        return self._async_queue("collapse", "view_method", idx)

    def set_depth(self, idx):
        return self._async_queue("set_depth", "view_method", idx)

    def column_paths(self):
        return self._async_queue("column_paths", "view_method")

    def schema(self, as_string=False):
        return self._async_queue("schema", "view_method", as_string=as_string)

    def expression_schema(self, as_string=False):
        return self._async_queue(
            "expression_schema", "view_method", as_string=as_string
        )

    def on_update(self, callback, mode=None):
        return self._subscribe("on_update", "view_method", callback, mode=mode)

    def remove_update(self, callback):
        return self._unsubscribe("remove_update", "view_method", callback)

    def on_delete(self, callback):
        return self._subscribe("on_delete", "view_method", callback)

    def remove_delete(self, callback):
        return self._unsubscribe("remove_delete", "view_method", callback)

    def delete(self):
        return self._async_queue("delete", "view_method")

    def to_arrow(self, **kwargs):
        return self._async_queue("to_arrow", "view_method", **kwargs)

    def to_records(self, **kwargs):
        return self._async_queue("to_records", "view_method", **kwargs)

    def to_dict(self, **kwargs):
        return self._async_queue("to_dict", "view_method", **kwargs)

    def to_numpy(self, **kwargs):
        return self._async_queue("to_numpy", "view_method", **kwargs)

    def to_df(self, **kwargs):
        return self._async_queue("to_df", "view_method", **kwargs)

    def to_csv(self, **kwargs):
        return self._async_queue("to_csv", "view_method", **kwargs)

    def to_json(self, **kwargs):
        return self._async_queue("to_json", "view_method", **kwargs)

    def to_columns(self, **kwargs):
        return self._async_queue("to_columns", "view_method", **kwargs)



################################################################################
#
# Copyright (c) 2019, the Perspective Authors.
#
# This file is part of the Perspective library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#

from .widget import PerspectiveWidget

__all__ = ["PerspectiveWidget"]
